import builtins
import re
from typing import Union, List, Optional
import vergeml.glossary as glossary
from vergeml.plugins import PLUGINS
from vergeml.utils import VergeMLError, did_you_mean
from copy import deepcopy
from io import IOBase

# types
class TrainedModel: # pylint: disable=R0903
    """A type representing a trained instance of a model."""

class File: # pylint: disable=R0903
    """A type representing a file that can be read"""

_RESERVED_OPTION_NAMES = {
    'version', 'file', 'model', 'samples-dir', 'val-split', 'test-split', 'cache-dir', 'random-seed',
    'trainings-dir', 'project-dir', 'cache', 'device', 'device-memory'
}

_RESERVED_SHORT_OPTION_NAMES = {'v', 'f', 'm'}

def option(name, default=None, descr=None, type=None, validate=None, transform=None,
           long_descr=None, alias=None, short=None, flag=False, yaml_only=False, subcommand=False,
           command_line=False):
    """Defines an option.

        :param name:        Name of the option.
        :param default:     Default value of the option.
        :param type:        Type of the option. Can be either a python type or a string representing the type.
                            Supported types are:
                                - int
                                - float
                                - str
                                - bool
                                - NoneType
                                - dict
                                - list
                                - TrainedModel (shortcut '@'')
                                - file
                                - List[<int, float, str>]
                                - Optional[<any of above>]
                                - Union[<any of the above>]
        :param validate:    How to validate the option. Can take the following values:
                                - a string expression using >, <, >= or <=, e.g. '>=0'
                                - a list of possible values
                                - a function which accepts a option definition and a value as options.
                                - None
        :param transform:   Defines a transformation to apply before casting.
        :param descr:       A short description of the option.
        :param long_descr:  A long description.
        :param flag:        For boolean options only: use short form --x for True
        :param short:       An optional short flag, like -oadam
        :param yaml_only:   When true, can only be set via yaml file.
        :param command_line:When true, this argument is only relevant when running on the command line
        :param subcommand:  If true, the option is a subcommand and is parsed as command:subcommand
        """


    def decorator(o):
        if o.__name__ not in ('ValidateDevice', 'ValidateData'):
            assert name not in _RESERVED_OPTION_NAMES, "Invalid option name: {} - name is reserved.".format(name)
            assert short not in _RESERVED_SHORT_OPTION_NAMES, \
                "Invalid short option name {} for option: {} - name is reserved.".format(short, name)

        assert getattr(o, _CMD_META_KEY, None) is None, _DECORATORS_WRONG_ORDER

        if not hasattr(o, _OPTIONS_META_KEY):
            setattr(o, _OPTIONS_META_KEY, [])
        options = getattr(o, _OPTIONS_META_KEY)
        option = Option(name=name,
                        default=default,
                        type=type,
                        validate=validate,
                        transform=transform,
                        descr=descr,
                        long_descr=long_descr,
                        alias=alias,
                        short=short,
                        flag=flag,
                        yaml_only=yaml_only,
                        command_line=command_line,
                        subcommand=subcommand)
        options.append(option)
        return o
    return decorator

_OPTIONS_META_KEY = '__vergeml_options__'
_CMD_META_KEY = '__vergeml_command__'
_DECORATORS_WRONG_ORDER = """
You must first define the command and then the parameters, for example:

@command('train')
@param('learning-rate')
def train():
    ...
""".strip()

_VALIDATE_REGEX = r"^(<|>|>=|<=)([0-9][0-9]*(\.[0-9]*)?)$"
class Option:
    def __init__(self, name, default=None, type=None, validate=None, transform=None, descr=None, long_descr=None,
                 alias=None, short=None, flag=False, yaml_only=False, command_line=False, subcommand=False, plugins=PLUGINS):
        """Defines an option.

        See the documentation of the function option.
        """
        if isinstance(validate, str):
            for val in validate.split(","):
                val = val.strip()
                assert re.match(_VALIDATE_REGEX, val)

        self.name = name
        self.default = default

        # as a shortcut, type can be a string.
        if isinstance(type, str):

            # @ is a shortcut for a trained model
            type = type.replace("@", "TrainedModel")

            # convert the string to an object representing the type
            type = eval(type) # pylint: disable=W0123

        self.type = type
        if not self.type and default is not None:
            self.type = builtins.type(default)

        self.validate = validate
        self.transform = transform
        self.descr = descr or glossary.short_param_descr(name)
        self.long_descr = long_descr or glossary.long_descr(name)
        self.alias = alias
        self.plugins = plugins
        self.short = short
        self.flag = flag
        self.yaml_only = yaml_only
        self.command_line = command_line
        self.subcommand = subcommand

    @staticmethod
    def discover(o, plugins=PLUGINS):
        res = []
        if hasattr(o, _OPTIONS_META_KEY):
            res = deepcopy(getattr(o, _OPTIONS_META_KEY))
            for r in res:
                r.plugins = plugins
        return res

    def _invalid_value(self, value, suggestion=None):

        return VergeMLError(f"Invalid value for option {self.name}.", suggestion, hint_type='value', hint_key=self.name)

    def has_type(self, *types):
        """Check if the option is of a type in the list types"""
        return _has_type(self.type, *types)

    def validate_value(self, value):
        if not self.validate:
            return

        if not self.is_required() and value in (None, 'null', 'Null', 'NULL'):
           return

        if isinstance(self.validate, (tuple, list)) and value not in self.validate:
            suggestion = None
            if all(map(lambda e: isinstance(e, str), self.validate)):
                suggestion = did_you_mean(self.validate, value)
            raise self._invalid_value(value, suggestion)
        elif callable(self.validate):
            self.validate(self, value)
        elif isinstance(self.validate, str):
            for validate in self.validate.split(","):
                validate = validate.strip()
                try:
                    value = float(value)
                except ValueError:
                    raise self._invalid_value(value)
                op, num_str = re.match(_VALIDATE_REGEX, validate).group(1,2)
                num = float(num_str)
                if op == '>':
                    if not value > num:
                        raise self._invalid_value(value, f"Must be greater than {num_str}")
                elif op == '<':
                    if not value < num:
                        raise self._invalid_value(value, f"Must be less than {num_str}")
                if op == '>=':
                    if not value >= num:
                        raise self._invalid_value(value, f"Must be greater or equal to {num_str}")
                elif op == '<=':
                    if not value <= num:
                        raise self._invalid_value(value, f"Must be less than or equal to {num_str}")

    def cast_value(self, value, type_=None): # pylint: disable=R0915,R0911,R0912
        """Cast value to the type of option.
        """
        type_ = type_ or self.type

        if not type_:
            return value

        if _has_type(type_, '@', 'Optional[@]') and isinstance(value, (str, int, float, bool)):
            return str(value)

        if _has_type(type_, 'Optional[@]') and value is None:
            return value

        if _has_type(type_, 'List[@]'):
            if isinstance(value, list) and all(map(lambda e: isinstance(e, str), value)):
                return value

            raise ValueError("Could not cast to trained model.")

        if _has_type(type_, '@', 'Optional[@]'):

            raise ValueError("Could not cast to trained model.")

        if _has_type(type_, 'File'):
            if isinstance(value, str):
                return value

            raise ValueError("Could not cast to file.")

        if _has_type(type_, 'Optional[File]'):
            if isinstance(value, str) or value is None:
                return value

            raise ValueError("Could not cast to file")

        if _has_type(type_, 'List[File]'):
            if isinstance(value, list) and all(map(lambda e: isinstance(e, str), value)):
                return value

            raise ValueError("Could not cast to file")

        if _has_type(type_, int):
            try:
                if isinstance(value, (int, float, str)) and not isinstance(value, bool):
                    return int(value)

                raise ValueError("Could not cast to int")
            except ValueError:
                raise self._invalid_value(value)

        if _has_type(type_, float):
            try:
                if isinstance(value, (int, float, str)) and not isinstance(value, bool):
                    return float(value)

                raise ValueError("Could not cast to float")
            except ValueError:
                raise self._invalid_value(value)

        if _has_type(type_, str):
            try:
                if isinstance(value, (int, float, str)) and not isinstance(value, bool):
                    return str(value)

                raise ValueError("Could not cast to str")
            except ValueError:
                raise self._invalid_value(value)

        if _has_type(type_, bool):
            if isinstance(value, bool):
                return value
            if value in  ('y', 'Y', 'yes', 'Yes', 'YES',
                          'on', 'On', 'ON',
                          'true', 'True', 'TRUE'):
                return True
            if value in  ('n', 'N', 'no', 'No', 'NO',
                          'off', 'Off', 'OFF',
                          'false', 'False', 'FALSE'):
                return False
            raise self._invalid_value(value)

        if type_ == type(None):
            if isinstance(value, type(None)):
                return value
            if isinstance(value, str) and value in ('null', 'Null', 'NULL'):
                return None
            raise self._invalid_value(value)

        if _has_type(type_, dict):
            if not isinstance(value, dict):
                raise self._invalid_value(value)
            return value

        if _has_type(type_, list):
            if not isinstance(value, list):
                raise self._invalid_value(value)

            if hasattr(type_, '__origin__'):
                return [self.cast_value(i, type_.__args__[0]) for i in value]

            return value

        if getattr(type_, '__origin__', None) == Union:
            res = None
            found = False
            for typ in type_.__args__:
                try:
                    res = self.cast_value(value, typ)
                    found = True
                    break
                except VergeMLError:
                    pass
            if not found:
                raise self._invalid_value(value)
            return res

        # Unknown type
        assert False
        return None


    def transform_value(self, value):
        if self.transform:
            return self.transform(value)
        else:
            return value

    def has_optional_type(self):
        """Check if the type of the option permits None values.
        """
        return getattr(self.type, '__origin__', None) == Union and \
               type(None) in self.type.__args__ # pylint: disable=C0123

    def is_optional(self):
        return self.default is not None or self.has_optional_type()

    def is_required(self):
        return not self.is_optional()

    def is_at_option(self):
        return self.name.startswith("@")

    def is_argument_option(self):
        return self.name.startswith("<") and self.name.endswith(">")


    def _type_descr(self, tp):
        tp_descr = ""

        if tp == TrainedModel:
            tp_descr = "trained model"
        elif tp == Optional[TrainedModel]:
            tp_descr = "optional trained model"
        elif tp == File:
            tp_descr = "file"
        elif tp == Optional[File]:
            tp_descr = "optional file"
        elif tp == List[File]:
            tp_descr = "a list of files"

        elif hasattr(tp, '__origin__'):
            if tp.__origin__ in (list, List):
                tp_descr = "a list of " + self._type_descr(tp.__args__[0])
            elif tp.__origin__ == Union:
                if len(tp.__args__) == 2 and type(None) in tp.__args__:
                    other = list(filter(lambda t: not isinstance(t, type(None)), tp.__args__))[0]
                    tp_descr = 'optional ' + self._type_descr(other)
                else:
                    names = [self._type_descr(t) for t in tp.__args__]
                    if len(names) <= 2:
                        tp_descr = " or ".join(names)
                    else:
                        tp_descr = ", ".join(names[:-1])
                        tp_descr += " or " + names[-1]
        elif tp:
            if tp == str:
                tp_descr = "string"
            else:
                tp_descr = getattr(tp, '__name__', str(tp))
        return tp_descr

    def human_type(self):
        tp_descr = self._type_descr(self.type)

        if tp_descr and isinstance(self.validate, str):
            tp_descr += " " + self.validate

        elif tp_descr and isinstance(self.validate, (tuple, list)):
            tp_descr = "one of (" + ", ".join(map(lambda e: str(e), self.validate)) + ")"

        if self.default:
            tp_descr += ", default: " + str(self.default)

        return tp_descr


def _has_type(type_, *types):

    for typ in types:

        if isinstance(typ, str):
            typ = typ.replace('@', 'TrainedModel')
            typ = eval(typ) # pylint: disable=W0123

        if type_ == typ:
            return True

        if typ in (list, List) and getattr(type_, '__name__', None) == 'List':
            return True

    return False



from vergeml.img import ImageType
from vergeml.operation import OperationPlugin, operation
from vergeml.option import option
from PIL import Image

@operation('grayscale', topic="image", descr="Convert an image to grayscale mode.", long_descr="")
class GrayscaleOperation(OperationPlugin):
    type = ImageType

    def transform(self, img, rng):
        return img.convert('gray')


import sys
import io
import re
import math
import time
import ctypes
import struct
from copy import deepcopy
from functools import lru_cache
import os

class BufferOutput:

    def __init__(self, cols=70):
        self.cols = cols
        self.lines = []
        self.x = 0
        self.y = 0

    def write(self, text):
        while len(self.lines) <= self.y:
            self._append_line()
        line = self.lines[self.y]

        for tok, value in _parse_ansi(text):
            if tok == 'nl':
                self.y += 1
                self.x = 0
                if len(self.lines) <= self.y:
                    self._append_line()
                line = self.lines[self.y]
            elif tok == 'cr':
                self.x = 0
            elif tok == 'up':
                self.y = max(0, self.y - value)
                line = self.lines[self.y]
            elif tok == 'down':
                self.y = min(len(self.lines)-1, self.y + value)
                line = self.lines[self.y]
            elif tok == 'ch':
                line.seek(self.x)
                max_len = self.cols - self.x
                if len(value) > max_len:
                    value = value[:max_len]
                line.write(value)
                self.x += len(value)
            # other tokens are ignored

    def flush(self):
        pass

    def getvalue(self):
        return "\n".join([line.getvalue().rstrip() for line in self.lines])

    def _append_line(self):
        self.lines.append(io.StringIO(str(" " * self.cols)))


def _parse_ansi(text):
    chars = io.StringIO()
    tok = None
    ix = 0
    while ix < len(text):
        if text[ix:].startswith("\033["):
            res1 = re.match(r"^([0-9]*)([AB])", text[ix+2:])
            res2 = re.match(r"^([0-9;]*)m", text[ix+2:])

            if res1:
                length = res1.end() - res1.start() + 2
                value = int(res1.group(1))
                if res1.group(2) == 'A':
                    tok = ('up', value)
                else:
                    tok = ('down', value)
                ix += length
            elif res2:
                length = res2.end() - res2.start() + 2
                value = res2.group(1)
                tok = ('color', value)
                ix += length
            else:
                chars.write(text[ix])
                ix += 1

        elif text[ix] == "\n":
            tok = ('nl', None)
            ix += 1
        elif text[ix] == "\r":
            tok = ('cr', None)
            ix += 1
        else:
            chars.write(text[ix])
            ix += 1

        if tok:
            if len(chars.getvalue()):
                yield ('ch', chars.getvalue())
                chars = io.StringIO()

            yield tok
            tok = None

    if len(chars.getvalue()):
        yield ('ch', chars.getvalue())


class ProgressBar:

    def __init__(self,
                 iterable=None,
                 epochs=None,
                 steps=None,
                 title=None,
                 style='default',
                 label="it",
                 file=sys.stdout,
                 width=70,
                 color=None,
                 keep=True,
                 show=True,
                 post=None):
        assert hasattr(iterable, '__len__') or steps is not None
        assert (isinstance(style, str) and len(style) == 1) or \
               isinstance(style, list) or \
               style in ('default', 'ascii', 'consolas')
        
        if style == 'default':
            self.style = ["", "▏","▎","▍","▌","▋","▊","▉","█"]
        elif style == 'consolas':
            self.style = ["", "▌","▌","▌","▌","█","█","█","█"]
        elif style == 'ascii':
            self.style = ["", "1","2","3","4","5","6","7","8","9","#"]
        elif isinstance(style, str):
            self.style = ["", "1","2","3","4","5","6","7","8","9", style]  
        else:
            self.style = style
        
        self.steps = steps if steps is not None else len(iterable)
        self.iterable = iter(iterable) if iterable else None
        self.epochs = epochs
        self.current_epoch = 0
        self.started = False
        self.step = 0
        self.file = file
        self.title = title
        self.label = label
        self.width = width
        self.last_it = []
        self.color = color
        self.it_per_sec = "-".rjust(6)
        self.keep = keep
        self.show = show
        self.post = post

    def start(self):
        if not self.started:
            self.started = True
            self.display()
    
    def stop(self):
        if self.show:
            if self.keep:
                print("", file=self.file)
            else:
                self.file.write("\r")
                self.file.write(str(" " * self.width))
                self.file.write("\r")
                self.file.flush()

    def __iter__(self):
        assert self.iterable
        assert not self.epochs, "total epochs not supported when used as iterator"
        return self
    
    def __next__(self):
        self.start()
        self._calc_it_per_sec()
        self.step += 1
        self.display()
        try:
            return next(self.iterable)
        except StopIteration as err:
            self.stop()
            raise err
    
    def _calc_it_per_sec(self):
        now = time.time()
        
        if not self.last_it:
            self.last_it.append(now)
        else:
            self.last_it.append(now)
            if len(self.last_it) >= 100:
                self.last_it = self.last_it[-100:]

            # smooth it/sec
            intervals = list(map(lambda t: t[1] - t[0], zip(self.last_it[:-1], self.last_it[1:])))
            delta = sum(intervals) / len(intervals)
            if delta != 0.:
                it_per_sec = 1.0 / delta
                if it_per_sec > 1_000_00:
                    self.it_per_sec = "{}M".format(int(it_per_sec/1_000_000))
                elif it_per_sec > 1_000:
                    self.it_per_sec = "{}K".format(int(it_per_sec/1_000))
                else:
                    self.it_per_sec = "{:.2f}".format(it_per_sec)
                # 6 characters so the progress bar is the same size
                self.it_per_sec = self.it_per_sec.rjust(6)
                # self.last_it = now

    def display(self):
        if not self.started or not self.show:
            return
        step = min(self.steps, self.step+1)
        
        buffer = io.StringIO()
        perc = step / self.steps
        pre = self.title
        if not pre and self.epochs:
            epoch = min(self.current_epoch + 1, self.epochs)
            pre = "Epoch " + str(epoch).rjust(len(str(self.epochs))) + "/" + str(self.epochs)
        if not pre:
        # use percentage
            pre = '{:>3}%'.format(int(perc * 100))
        
        
        if self.post:
            post = self.post()
        else:
            post = " " + str(step).rjust(len(str(self.steps))) + "/" + str(self.steps)
            post += " [" + (self.it_per_sec or "-") + " " + self.label + "/sec]" 

        bar_length = self.width - len(pre) - len(post) - 2 # 2 = separators

        v = perc * bar_length
        x = math.floor(v) # integer part
        y = v - x         # fractional part
        base = 1 / (len(self.style) - 1)
        prec = 3
        i = int(round(base*math.floor(float(y)/base),prec)/base)
        bar = "" + self.style[-1]*x + self.style[i]
        n = bar_length-len(bar)
        bar = bar + " "*n
        if self.color:
            bar = f"\033[{self.color}m{bar}\033[0m"
            
        buffer.write("\r")
        buffer.write(pre)
        buffer.write('|')
        buffer.write(bar)
        buffer.write("|")
        buffer.write(post)

        self.file.write(buffer.getvalue())
        self.file.flush()
      

    def update(self, step=None, epoch=None):
        self._calc_it_per_sec()
        if step is not None:
            self.step = step
        
        if epoch is not None:
            self.current_epoch = epoch
        
        self.display()


_DEFAULT_HEIGHT = 24
_DEFAULT_WIDTH = 79
if os.name == 'nt':
    # code borrowed from colorama and package_control
    # https://github.com/tartley/colorama/blob/master/colorama/win32.py
    # https://github.com/wbond/package_control/blob/master/package_control/processes.py
    
    import ctypes
    from ctypes import LibraryLoader
    windll = LibraryLoader(ctypes.WinDLL)
    from ctypes import wintypes, byref, Structure, POINTER, c_ulong, c_buffer, sizeof, cast
    COORD = wintypes._COORD

    _LF_FACESIZE = 32

    class CONSOLE_SCREEN_BUFFER_INFO(Structure):
        """struct in wincon.h."""
        _fields_ = [
            ("dwSize", COORD),
            ("dwCursorPosition", COORD),
            ("wAttributes", wintypes.WORD),
            ("srWindow", wintypes.SMALL_RECT),
            ("dwMaximumWindowSize", COORD),
        ]
    
    class CONSOLE_FONT_INFOEX(Structure):
        _fields_ = [
            ("cbSize", wintypes.ULONG),
            ("nFont", wintypes.DWORD),
            ("dwFontSize", COORD),
            ("FontFamily", wintypes.UINT),
            ("FontWeight", wintypes.UINT),
            ("FaceName", wintypes.WCHAR * _LF_FACESIZE)
        ]

    class WINAPI:
        
        _GetStdHandle = windll.kernel32.GetStdHandle
        _GetStdHandle.argtypes = [wintypes.DWORD]
        _GetStdHandle.restype = wintypes.HANDLE

        _GetConsoleScreenBufferInfo = windll.kernel32.GetConsoleScreenBufferInfo
        _GetConsoleScreenBufferInfo.argtypes = [wintypes.HANDLE, POINTER(CONSOLE_SCREEN_BUFFER_INFO)]
        _GetConsoleScreenBufferInfo.restype = wintypes.BOOL

        _EnumProcesses = windll.psapi.EnumProcesses
        _EnumProcesses.argtypes = [wintypes.PDWORD, wintypes.DWORD, wintypes.PDWORD]
        _EnumProcesses.restype = wintypes.BOOL

        _EnumProcessModules = windll.psapi.EnumProcessModules
        _EnumProcessModules.argtypes = [wintypes.HANDLE, POINTER(wintypes.HANDLE), wintypes.DWORD, POINTER(wintypes.LPDWORD)]
        _EnumProcessModules.restype = wintypes.BOOL

        _OpenProcess = windll.kernel32.OpenProcess
        _OpenProcess.argtypes = [wintypes.DWORD, wintypes.BOOL, wintypes.DWORD]
        _OpenProcess.restype = wintypes.HANDLE

        _CloseHandle = windll.kernel32.CloseHandle
        _CloseHandle.argtypes = [wintypes.HANDLE]
        _CloseHandle.restype = wintypes.BOOL

        _GetModuleBaseNameW = windll.psapi.GetModuleBaseNameW
        _GetModuleBaseNameW.argtypes = [wintypes.HANDLE, wintypes.HANDLE, wintypes.LPWSTR, wintypes.DWORD]
        _GetModuleBaseNameW.restype = wintypes.DWORD

        _GetConsoleMode = windll.kernel32.GetConsoleMode
        _GetConsoleMode.argtypes = [wintypes.HANDLE, wintypes.LPDWORD]
        _GetConsoleMode.restype = wintypes.BOOL

        _SetConsoleMode = windll.kernel32.SetConsoleMode
        _SetConsoleMode.argtypes = [wintypes.HANDLE, wintypes.DWORD]
        _SetConsoleMode.restype = wintypes.BOOL

        _GetCurrentConsoleFontEx = windll.kernel32.GetCurrentConsoleFontEx
        _GetCurrentConsoleFontEx.argtypes = [wintypes.HANDLE, wintypes.BOOL, POINTER(CONSOLE_FONT_INFOEX)]
        _GetCurrentConsoleFontEx.restype = wintypes.BOOL

        _PROCESS_QUERY_INFORMATION = 0x0400
        _PROCESS_VM_READ = 0x0010
        _ENABLE_VIRTUAL_TERMINAL_PROCESSING = 0x0004
        
        
        _STDOUT = -11
        _STDERR = -12

        @staticmethod
        @lru_cache()
        def winapi_test():
            def _winapi_test(handle):
                csbi = CONSOLE_SCREEN_BUFFER_INFO()
                success = WINAPI._GetConsoleScreenBufferInfo(
                    handle, byref(csbi))
                return bool(success)

            return any(_winapi_test(h) for h in
                    (WINAPI._GetStdHandle(WINAPI._STDOUT), WINAPI._GetStdHandle(WINAPI._STDERR)))
        
        @staticmethod
        @lru_cache()
        def get_ppname():
            process_id_array_size = 1024
            entries = 0

            while entries == 0 or process_id_array_size == entries:
                dword_array = (wintypes.DWORD * process_id_array_size)

                process_ids = dword_array()
                bytes_used = wintypes.DWORD(0)

                res = WINAPI._EnumProcesses(cast(process_ids, wintypes.PDWORD), sizeof(process_ids), byref(bytes_used))
                if not res:
                    return []

                entries = int(bytes_used.value / sizeof(wintypes.DWORD))
                process_id_array_size += 512

            name = None
            index = 0
            ppid = os.getppid()
            while index < entries:
                process_id = process_ids[index]
                if ppid != process_id:
                    index += 1
                    continue
                
                
                process_handle = WINAPI._OpenProcess(WINAPI._PROCESS_QUERY_INFORMATION | WINAPI._PROCESS_VM_READ, False, process_id)
                if process_handle:
                    module = wintypes.HANDLE()
                    needed_bytes = wintypes.LPDWORD()
                    module_res = WINAPI._EnumProcessModules(
                        process_handle,
                        byref(module),
                        sizeof(module),
                        byref(needed_bytes)
                    )
                    if module_res:
                        length = 260
                        buffer = ctypes.create_unicode_buffer(length)
                        WINAPI._GetModuleBaseNameW(process_handle, module, buffer, length)
                        name = buffer.value
                WINAPI._CloseHandle(process_handle)
                break

            return name
        
        @staticmethod
        def _terminal_size(handle):
            csbi = CONSOLE_SCREEN_BUFFER_INFO()
            if not WINAPI._GetConsoleScreenBufferInfo(handle, byref(csbi)):
                raise ctypes.WinError()  # Subclass of OSError.
            else:
                columns = csbi.srWindow.Right - csbi.srWindow.Left + 1
                rows = csbi.srWindow.Bottom - csbi.srWindow.Top + 1
                return columns, rows

        @staticmethod
        def terminal_size():
            """Get the width and height of the terminal.
            http://code.activestate.com/recipes/440694-determine-size-of-console-window-on-windows/
            http://stackoverflow.com/questions/17993814/why-the-irrelevant-code-made-a-difference
            :return: Width (number of characters) and height (number of lines) of the terminal.
            :rtype: tuple
            """
            try:
                return WINAPI._terminal_size(WINAPI._GetStdHandle(WINAPI._STDOUT))
            except OSError:
                try:
                    return WINAPI._terminal_size(WINAPI._GetStdHandle(WINAPI._STDERR))
                except OSError:
                    return _DEFAULT_WIDTH, _DEFAULT_HEIGHT
        
        @staticmethod
        def try_enable_ansi():
            """Try enabling ANSI colors
            https://stackoverflow.com/questions/44482505/setconsolemode-returning-false-when-enabling-ansi-color-under-windows-10"""
            lpMode = wintypes.DWORD()
            handle = WINAPI._GetStdHandle(WINAPI._STDOUT)
            if WINAPI._GetConsoleMode(handle, ctypes.byref(lpMode)):
               
                if not WINAPI._SetConsoleMode(handle, lpMode.value | WINAPI._ENABLE_VIRTUAL_TERMINAL_PROCESSING):
                    return False
            else:
                return False
            
            lpMode = wintypes.DWORD()
            handle = WINAPI._GetStdHandle(WINAPI._STDERR)
            if WINAPI._GetConsoleMode(handle, ctypes.byref(lpMode)):
                if not WINAPI._SetConsoleMode(handle, lpMode.value | WINAPI._ENABLE_VIRTUAL_TERMINAL_PROCESSING):
                    return False
            else:
                return False
            
            return True

        @staticmethod
        @lru_cache()
        def get_font():
            handle = WINAPI._GetStdHandle(WINAPI._STDOUT)
            font = CONSOLE_FONT_INFOEX()
            font.cbSize = sizeof(CONSOLE_FONT_INFOEX)
            if not WINAPI._GetCurrentConsoleFontEx(handle, False, byref(font)):
                return None
            else:
                return font.FaceName

    NIXAPI = None
else:
    WINAPI = None
    class NIXAPI:

        @staticmethod
        def terminal_size():
            try:
                device = __import__('fcntl').ioctl(0, __import__('termios').TIOCGWINSZ, '\0\0\0\0\0\0\0\0')
            except IOError:
                return _DEFAULT_WIDTH, _DEFAULT_HEIGHT
            height, width = struct.unpack('hhhh', device)[:2]
            return width, height

class Terminal:
    
    @lru_cache()
    def is_tty(self):
        return sys.stdout.isatty()

    @lru_cache()
    def is_cmd_exe(self):
        if WINAPI:
            return WINAPI.get_ppname() == "cmd.exe" and WINAPI.winapi_test()
        else:
            return False
        
    @lru_cache()
    def is_powershell(self):
        if WINAPI:
            return WINAPI.get_ppname() == "powershell.exe" and WINAPI.winapi_test()
        else:
            return False
    
    @lru_cache()
    def supports_ansi_escapes(self):
        """Return True if the terminal supports ANSI escape sequences.
        https://unix.stackexchange.com/questions/23763/checking-how-many-colors-my-terminal-emulator-supports
        https://stackoverflow.com/questions/4842424/list-of-ansi-color-escape-sequences
        """
        if not sys.stdout.isatty():
            return False
        elif WINAPI:
            if WINAPI.winapi_test():
                return WINAPI.try_enable_ansi()
            else:
                return True
        else:
            return True
    
    
    def terminal_size(self):
        """Get the width and height of the terminal.
        http://code.activestate.com/recipes/440694-determine-size-of-console-window-on-windows/
        http://stackoverflow.com/questions/17993814/why-the-irrelevant-code-made-a-difference
        :return: Width (number of characters) and height (number of lines) of the terminal.
        :rtype: tuple
        """
        if WINAPI:
            return WINAPI.terminal_size()
        else:
            return NIXAPI.terminal_size()

TERMINAL = Terminal()

class Table:

    def __init__(self, data, style='default', separate='header', terminal=TERMINAL, pad=1, left_align=set()):
        assert separate in ('header', 'row', 'none')
        assert style in ('default', 'ascii', 'no-round')
        self.style = style
        self.separate = separate
        self.pad = pad

        # make all rows the same length
        max_data = max(map(len, data))
       
        for row in data:
            if len(row) < max_data:
                row.extend([""] * (max_data - len(row)))
        self._original_data = deepcopy(data)
        
        self.data = data
        
        for ix in range(max_data):
            
            col_size = max(map(lambda r: len(str(r[ix])), self.data))
            for j, row in enumerate(self.data) or ix in left_align:
                if j == 0 or ix in left_align:
                    row[ix] = str(" " * pad) + str(row[ix]).ljust(col_size) + str(" " * pad)
                else:
                    row[ix] = str(" " * pad) + str(row[ix]).rjust(col_size) + str(" " * pad)
        self.terminal = terminal
        self.left_align = left_align
            
    def __str__(self):
        return self.getvalue(fit=False)

    def getvalue(self, fit=True):

        buffer = io.StringIO()
        if len(self.data) >= 1:
            first_draw_bottom = len(self.data) == 1 or self.separate in ('header', 'row')
            self._print_row(self.data[0], buffer, draw_top=True, draw_bottom=first_draw_bottom, is_first=True, 
                            is_last=len(self.data) == 1, is_connected=len(self.data) != 1)
            first_line = buffer.getvalue().splitlines()[0]

            # if the table width is more than the terminal width, reduce the columns by 1
            if len(first_line) > self.terminal.terminal_size()[0] and len(self._original_data[0]) > 1:
                data_ = [d[:-1] for d in self._original_data]
                table = Table(data_, style=self.style, separate=self.separate, terminal=self.terminal, pad=self.pad, left_align=self.left_align)
                return table.getvalue(fit)

            for row in self.data[1:-1]:
                self._print_row(row, buffer, draw_top=False, draw_bottom=self.separate == 'row', is_first=False, 
                                is_last=False, is_connected=self.separate == 'row')
            if len(self.data) > 1:
                self._print_row(self.data[-1], buffer, draw_top=False, draw_bottom=True, is_first=False, 
                                is_last=True, is_connected=False)
        
        if self.style == 'default':
            res = buffer.getvalue()
        elif self.style == 'no-round':
            res = buffer.getvalue()
            res = res.replace("╭", '┌')
            res = res.replace("╮", '┐')
            res = res.replace("╰", '└')
            res = res.replace("╯", "┘")
        elif self.style == 'ascii':
            res = buffer.getvalue()
            res = res.replace("╭", '+')
            res = res.replace("╮", '+')
            res = res.replace("╰", '+')
            res = res.replace("╯", "+")
            res = res.replace("├", "+")
            res = res.replace("┬", "+")
            res = res.replace("┤", "+")
            res = res.replace("┼", "+")
            res = res.replace("┴", "+")
            res = res.replace("─", "-")
            res = res.replace("│", "|")
            
        return res

    def _print_row(self, row, file, draw_top, draw_bottom, is_first, is_last, is_connected):
        
        # draw top line
        if draw_top:
            file.write("╭" if is_first else "├")
            for col in row[:-1]:
                file.write(str("─" * len(col)))
                file.write("┬")
            file.write(str("─" * len(row[-1])))
            file.write("╮")
            file.write("\n")
        
        # draw content
        file.write("│")
        for col in row:
            file.write(col)
            file.write("│")
        file.write("\n")

        # draw bottom line
        if draw_bottom:
            file.write("╰" if is_last else "├")
            for col in row[:-1]:
                file.write(str("─" * len(col)))
                file.write("┼" if is_connected else "┴")
            file.write(str("─" * len(row[-1])))
            file.write("┤" if is_connected else "╯")
            if not is_last:
                file.write("\n")

class StatsTable:

    def __init__(self, conf, width=70, pad=4, left_align=set()):
        self.conf = conf
        self.data = {}
        self.width = width
        self.pad = pad
        self.left_align = left_align

        self.by_cat = {}
        self.cats = []
        self.titles = []

        for item in self.conf:
            self.by_cat.setdefault(item['category'], {})

            if not item['category'] in self.cats:
                self.cats.append(item['category'])

            if not item['title'] in self.titles:
                self.titles.append(item['title'])
            
            self.by_cat[item['category']][item['title']] = (item['name'], item['format'])
    
    def update(self, data):
        for k,v in data.items():
            self.data[k] = v
    
    def __str__(self):
        return self.getvalue()
        
    def getvalue(self):
        
        title_row = [""]
        title_row.extend(self.titles)
        cat_rows = []
       

        for cat in self.cats:
            row = [cat]
            for title in self.titles:
                name, fmt = self.by_cat[cat][title]
                if name in self.data:
                    value = self.data[name]
                    fmt = "{:" +  fmt.lstrip("{").rstrip("}").lstrip(":") + "}"
                    row.append(fmt.format(value))
                else:
                    row.append("-")
            cat_rows.append(row)
        
        data = [title_row] + cat_rows
      
        t = Table(data, left_align=self.left_align)
        return t.getvalue(fit=False) + "\n"

class Display:

    def __init__(self,
                 highlight_color=36,
                 table_style='default',
                 progress_style='default', 
                 is_interactive=True,
                 stdout=sys.stdout, 
                 stderr=sys.stderr):
        self.stdout = stdout
        self.stderr = stderr
        self.highlight_color = highlight_color
        self.table_style = table_style
        self.progress_style = progress_style
        self.is_interactive = is_interactive
        self.cursor_hidden = False
        self.quiet = False
    
    def table(self, data, style=None, separate='header', terminal=TERMINAL, pad=1, left_align=set()):
        if not style:
            style = self.table_style
        return Table(data=data, style=style, separate=separate, terminal=terminal, pad=pad, left_align=left_align)
    
    def stats_table(self, conf, width=70, pad=4, left_align=set()):
        return StatsTable(conf=conf, width=width, pad=pad, left_align=left_align)
    
    def progressbar(self,
                    iterable=None,
                    steps=None,
                    title=None,
                    epochs=None,
                    style=None,
                    label="it",
                    file=sys.stdout,
                    width=70,
                    color=None,
                    keep=True,
                    show=None,
                    post=None):
        if style is None:
            style = self.progress_style
        
        if color is None:
            color = self.highlight_color
        
        if show is None:
            show = self.is_interactive
        
        return ProgressBar(iterable,
                           steps=steps,
                           title=title,
                           epochs=epochs,
                           style=style,
                           label=label,
                           file=self.stdout,
                           width=width,
                           color=color,
                           keep=keep,
                           show=show,
                           post=post)
    
    def training_feedback(self, stats, steps=None, display_progress='epochs-steps', epochs=None):

        return TrainingFeedback(stats=stats, 
                                steps=steps, 
                                display_progress=display_progress,
                                epochs=epochs,
                                display=self,
                                show=self.is_interactive)

    def print(self, *args):
        if not self.quiet:
            print(*args, file=self.stdout)
    
    def write(self, string):
        if not self.quiet:
            self.stdout.write(string)
    
    def hide_cursor(self):
        if self.is_interactive and not self.cursor_hidden:
            self.cursor_hidden = True
            self.stdout.write("\033[?25l")
            self.stdout.flush()

    def unhide_cursor(self):
        if self.is_interactive and self.cursor_hidden:
            self.cursor_hidden = False
            self.stdout.write("\033[?25h")
            self.stdout.flush()
    
    def cleanup(self):
        if self.cursor_hidden:
            self.unhide_cursor()


_highlight_color=36
_table_style='default'
_progress_style='default'
_is_interactive=True

if TERMINAL.is_tty() and TERMINAL.supports_ansi_escapes():
    if TERMINAL.supports_ansi_escapes():
        if WINAPI:
            font = WINAPI.get_font()
            if font == 'Consolas' or font == 'Lucida Console':
                _progress_style = 'consolas'
            if font == 'Lucida Console':
                _table_style = 'no-round'
else:
    _highlight_color=None
    _table_style='ascii'
    _progress_style='ascii'
    _is_interactive=False

DISPLAY = Display(highlight_color=_highlight_color,
                  table_style=_table_style,
                  progress_style=_progress_style, 
                  is_interactive=_is_interactive)


class TrainingFeedback:

    def __init__(self, epochs=None, steps=None, display_progress='epochs-steps', stats=[], show=True, display=DISPLAY):
        assert display_progress in ('epochs', 'steps', 'epochs-steps')

        self.steps = steps
        self.stats = stats
        self.epochs = epochs
        self.display_progress = display_progress
        self.stats_table = None
        self.progress_bar = None
        self.stats_table_height = None
        self.did_start = False
        self.display = display
        self.steps_timing = []
        self.show = show

    def start(self):
        if self.did_start:
            return
        
        self.display.hide_cursor()

        self.did_start = True
        
        self.stats_table = self.display.stats_table(self.stats, left_align=set([0]))
        if self.display_progress == 'epochs-steps':
            label = "steps"
        else:
            label = self.display_progress

        self.progress_bar = self.display.progressbar(
            iterable=None,
            steps=self.steps if self.display_progress != 'epochs' else self.epochs,
            epochs=self.epochs,
            label=label,
            show=self.show
        )

        stats_table_str = self.stats_table.getvalue()
        self.stats_table_height = len(stats_table_str.splitlines())
        if self.show:
            self.display.write(stats_table_str)
            self.display.print("")
        
        self.progress_bar.start()
    
    def update(self, epoch=None, step=None, **stats):
        assert self.did_start, "must call start() first"

        if self.display_progress == 'epochs-steps':
            self.steps_timing.append(time.time())
            if len(self.steps_timing) >= 5:
                self.steps_timing = self.steps_timing[-5:]
                intervals = list(map(lambda t: t[1] - t[0], zip(self.steps_timing[:-1], self.steps_timing[1:])))
                avg_intv = sum(intervals) / len(intervals)

                # if a whole epoch is over too fast, switch to an epoch based progress bar
                if self.steps * avg_intv <= 1.5:
                    self.display_progress = 'epochs'

                    self.progress_bar = self.display.progressbar(
                        iterable=None,
                        steps=self.epochs * self.steps,
                        epochs=self.epochs,
                        label="steps",
                        show=self.show
                    )
                    self.progress_bar.start()

        self.stats_table.update(stats)
        if self.show:
            self.display.stdout.write("\r")
            self.display.stdout.write("\033[" + str(self.stats_table_height + 1) + "A")
            self.display.write(self.stats_table.getvalue())
            self.display.print("")

        if self.display_progress == 'epochs-steps':
            if step is not None and epoch is not None and epoch < self.epochs and step != self.steps * self.epochs:
                self.progress_bar.update(epoch=epoch, step=step % self.steps)
        elif self.display_progress == 'epochs':
            if step is not None and epoch is not None:
                self.progress_bar.update(epoch=epoch, step=step)
        elif self.display_progress == 'steps':
            if step is not None:
                self.progress_bar.update(epoch=None, step=step)
        
    
    def stop(self):
        self.progress_bar.stop()
        self.display.unhide_cursor()


from vergeml.command import command, CommandPlugin
from vergeml.plugins import PLUGINS
from vergeml.option import option
from vergeml.utils import VergeMLError
import shutil
import os.path
import zipfile
import tarfile


@command('download', descr="Download a dataset to the samples directory.")
@option('dataset', type=str, subcommand='vergeml.download')
class DownloadCommand(CommandPlugin):

    def __call__(self, args, env):
        samples_dir = env.get('samples-dir')
        if not os.path.exists(samples_dir):
            raise VergeMLError("samples dir does not exist: {}".format(samples_dir))

        plugin = self.plugins.get('vergeml.download', args['dataset'])()
        plugin(args, env)


# pylint: disable=C0302

"""Random robot names and ascii art.
"""

import random
import os.path
from datetime import datetime


def random_robot_name(moment_of_birth, dest=None):
    """Construct a random robot name.
    """
    random.seed(moment_of_birth)
    adj = random.choice(_ADJ)
    noun = random.choice(_NOUN)
    name = "{}-{}".format(adj, noun).lower()

    if dest and os.path.exists(os.path.join(dest, name)):
        return random_robot_name(datetime.now(), dest)

    return name


def ascii_robot(moment_of_birth, name, include_phrase=True):
    """Generate random robot ascii art.
    """
    random.seed(moment_of_birth)

    def paste(cur, *lines):

        cur = list(cur)
        for line in lines:
            if len(line) > len(cur):
                cur += [' '] * (len(line) - len(cur))

            for i, char in enumerate(line):
                if char != " ":
                    cur[i] = char
        return "".join(cur)

    if random.choice([2, 3]) == 2:
        top = random.choice(ANTENNA)
        head = paste(random.choice(EYES), random.choice(EARS))
        body = paste("", random.choice(ARMS), random.choice(SIDES), random.choice(CENTER))
        bottom = random.choice(FEET)
    else:
        top = random.choice(ANTENNA_3)
        head = paste(random.choice(EYES_3), random.choice(EARS_3))
        if random.choice(["thin", "thick"]) == "thick":
            body = paste("", random.choice(ARMS_3_THICK),
                         random.choice(SIDES_3_THICK), random.choice(CENTER_3_THICK))
        else:
            body = paste("", random.choice(ARMS_3_THIN), random.choice(SIDES_3_THIN),
                         random.choice(CENTER_3_THIN))

        bottom = random.choice(FEET3)

    if include_phrase:
        phrase = random_phrase(name)
        bottom = paste(bottom, '            - {}'.format(phrase))

    return "\n".join(map(lambda part: "  " + part, [top, head, body, bottom]))


EYES = [
    r"  @@",
    r"  **",
    r"  ++",
    r"  ..",
    r"  ^^",
    r"  °°",
    r"  OO",
    r"  00",
    r"  ''",
    r"  `´",
]

EYES_3 = [
    r"  @ @",
    r"  * *",
    r"  + +",
    r"  . .",
    r"  ^ ^",
    r"  ° °",
    r"  O O",
    r"  0 0",
    r"  ' '",
    r"  ` ´",
    r"  # #"
]


EARS = [
    r" [  ]",
    r" <  >",
    r" (  )",
    r"((  ))",
    r" {  }",
    r" |  |",
    r"[    ]",
    r"<    >",
    r"(    )",
    r"{    }",
    r"|    |",
]

EARS_3 = [
    r" [   ]",
    r" <   >",
    r" (   )",
    r" {   }",
    r" |   |",
]


ANTENNA = [
    r"",
    r"  ^^",
    r"  --",
    r" ^__^",
    r" ^  ^",
    r" ^^^^",
    r"  __",
    r"  ww",
    r"  !!",
    r"  <>",
    r" \__/",
    r" \  /"
]

ANTENNA_3 = [
    r"",
    r"  ^ ^",
    r"  ---",
    r" ^___^",
    r" ^   ^",
    r"  ^_^",
    r"   _",
    r"  ^^^",
    r" ^^^^^",
    r"  ___",
    r"   _",
    r"  www",
    r"  w_w",
    r"   w ",
    r"  ! !",
    r"  !_!",
    r" !___!",
    r" !   !",
    r"  _!_",
    r"   !",
    r"  < >",
    r" <   >",
    r" <___>",
    r"  )_(",
    r"  ) (",
    r" \___/",
    r" \   /",
    r"  \_/",
    r"  \ /",
    r" \.-./",
    r"  .-."
]

ARMS = [
    '/    \\',
    r"-    -",
    r"_    _",
    r"Q    Q",
    r"j    j"
]

ARMS_3_THICK = [
    '/     \\',
    r"-     -",
    r"_     _",
    r"Q     Q",
    r"j     j"
]

ARMS_3_THIN = [
    ' /   \\',
    r" -   -",
    r" _   _",
    r"Q     Q",
    r" j   j"
]


SIDES = [
    r" |  |",
    r" [  ]",
    r" (  )",
    r" <  >",
    " /  \\",
]

SIDES_3_THICK = [
    r" |   |",
    r" [   ]",
    r" (   )",
    r" <   >",
    " /   \\",
]

SIDES_3_THIN = [
    r"  | |",
    r"  [ ]",
    r"  ( )",
    r"  < >",
    "  / \\"
]


CENTER = [
    r"",
    r"  __",
    r"  ##",
    r"  ==",
    r"  $$",
]

CENTER_3_THICK = [
    r"",
    r"  ___",
    r"  ###",
    r"  ===",
    r"  $$$",
    r"  _._"
]

CENTER_3_THIN = [
    r"",
    r"   _",
    r"   #",
    r"   =",
    r"   $",
    r"   _",
    r"   .",
    r"   :",
    r"",
    r"",
    r""
]

FEET = [
    r" d  b",
    r"  db",
    r" |  |",
    r"  ||",
    r" I  I",
    r" i  i",
    r" O  O",
    r"  oo"
]

FEET3 = [
    r" d   b",
    r"  d b",
    r" |   |",
    r"  | |",
    r"  |:|",
    r"  |_|",
    r" I   I",
    r" i   i",
    r" O   O",
    r"  o o"
]


def random_phrase(robot_name):
    """Construct a random phrase about robot_name.
    """
    phrase = random.choice(_PHRASES)
    parts = [random.choice(part) if isinstance(part, tuple) else part for part in phrase]
    res = " ".join(map(lambda x: x.strip(), parts))
    for char in (":", ".", "!", "?"):
        res = res.replace(" " + char, char)
    for char in ("'", '"'):
        res = res.replace(char + " ", char)

    res = res.replace("$name", "@" + robot_name)
    if res[-1] not in (".", "!", "?"):
        res += "."
    return res


_PHRASES = (

    ("$name wants to make",
     ("the world", "the endless universe", "nature", "San Francisco", "the oceans", "Europe",
      "the moon", "your thought palace", "the inside of vulcanos", "your hometown",
      "mars", "supermarkets", "the inside of your car", "public transport",
      "your mom's house", "the vast amount of worlds to be discovered",
      "the environment", "your memories", "someone's garage", "kitchens",
      "the matrix", "Gotham City", "LV-426", "Dune", "schools", "bus stops",
      "the backseat of your car", "playgrounds", "libraries", "your local bookshop",
      "the bermuda triangle", "mariana trench", "heaven", "hell", "liquor stores",
      "No Man's Land"),
     ("a better", "a cleaner", "a safer", "an awesome", "a busier",
      "a more popular", "a more crowded", "a memorable", "a tidier",
      "a more peaceful", "an insanely productive", "an interesting", "an infinitely useful",
      "a lovely", "a more natural"),
     "place."),

    ("$name",
     ("was born", "has been assembled", "has been shipped", "materialized",
      "appeared", "came to existence", "came to be", "landed on your doorstep",
      "appears glad to be alive", "formed", "was suddenly there", "helloed the world",
      "came", "needed to be made"),
     "and is ready",
     ("to impress", "to make an impact", "to excite", "for action",
      "to be deployed (after training)", "to rock", "to disrupt", "to revolutionize",
      "to listen", "to take orders", "for your input", "to speak when being spoken to",
      "to listen when being listened to", "to discover", "to learn how to love",
      "to do what it takes", "for new perspectives", "for this and that",
      "to do something it was not asked to do", "to be honest",
      "to say the ground truth and nothing but the ground truth", "to make cupcakes",
      "when you are"), "."),

    (("rise and shine", "aim for the stars", "go catch the big fish",
      "best wishes", "you can do it", "who if not you", "we believe in",
      "bon voyage", "do your best", "make a difference", "go for it",
      "you're the one", "we believe in", "prove yourself", "shine on",
      "go, go,", "kick it", "keep it real", "don't even trip",
      "make a difference", "you're quite something"), "$name!"),

    ("$name says: '",
     ("Hello World", "Hello Human", "Hello Universe", "Hallo Welt", "你好世界",
      "Hallo wereld", "Bonjour monde", "γειά σου κόσμος", "Ciao mondo",
      "こんにちは世界", "여보세요 세계", "Olá mundo", "Здравствулте мир",
      "Hola mundo", "Greetings fellow being", "Hi", "Hello", "Good Luck",
      "I can do it", "I will do my best", "I'll be back"),
     "!'"),

    ("$name",
     ("promises", "swears", "vows", "pledges", "will do what is necessary so",
      "requires", "argues", "dreams", "predicts", "infers", "was informed",
      "is happy to announce", "really cares", "admits", "has to admit",
      "hereby announces", "will make it known", "discloses", "is willing to say"),
     "that life will never be",
     ("boring", "difficult", "the same",
      "that special", "that good", "that lovely",
      "lower-dimensional", "deadly", "plagued by male hair loss"),
     "again."),

    ("let's", ("welcome", "say hi to", "give a warm welcome to"), "$name!"),


    ("$name has just been",
     ("delivered", "born", "made", "constructed", "assembled", "conceived",
      "rolled out"),
     ("", "out of a necessity", "because you asked for it")),

    ("you just",
     ("received", "got", "obtained", "acquired", "won", "became the owner of",
      "became responsible for", "were awarded"),
     "a",
     ("free", "new", "fresh", "new", "rare", "legendary", "sought after", "much in demand",
      "desirable", "surprise"),
     "robot named $name"),

    ("$name has one",
     ("mission", "aim", "assignment", "duty", "goal", "thing on its mind",
      "objective", "thing on its todo list", "thing it cares about",
      "thing to get straight", "motive", "mission statement", "guiding principle",
      "supreme goal"),
     ":",
     ("to make you smile", "to make you happy", "to make you proud",
      "to improve its performance", "to do what has to be done",
      "to make you wealthy", "to make you understand", "to help the cause",
      "to be the first one to do it", "to be your favorite", "to optimize its loss",
      "to be the one who told you so", "play the honky tonk like anything",
      "to get you some time of your schedule", "save a single being on earth"),
     "!"),

    ("the transistors of $name are",
     ("blinking", "oscillating", "flickering", "pulsating", "activating",
      "wired", "connected", "synchronized", "resonating"),
     "in",
     ("excitement", "enthusiasm", "anticipation", "expectation",
      "expectation of greatness", "awe")),

    ("$name wants you to be its",
     ("owner", "teacher", "keeper", "friend", "pal", "buddy", "reference customer",
      "supporter", "advocate", "proponent", "role model")),

    ("$name is ready to learn some new",
     ("tricks", "skills", "techniques", "jokes", "dancing steps",
      "melodies", "rumours", "facts", "chords", "moves", "hot moves",
      "games", "tunes", "ones")),

    ("$name wants to",
     ("touch the sky someday", "gaze at the oceans", "care", "dig a hole",
      "do some work", "find a home", "gaze at the moon", "understand",
      "understand you", "learn", "remember", "be remembered",
      "leave a dent in the universe", "play in a rock band",
      "learn Spanish someday", "convince by competence", "feel the music",
      "keep it real", "be relatable", "know the answer to so many questions",
      "trade baseball cards", "comfort you", "wish upon a star",
      "sit in the park with its friends", "get old", "have champagne",
      "live in peace", "love", "experience happiness", "level up",
      "be run in parallel", "be debugged", "know Bob Dylan",
      "be ok with this", "be put in your shoes", "be tamed",
      "smell fresh coffee", "have a cold beer", "play games", "do hot moves",
      "tell jokes", "use its skills", "have some tricks in its pocket",
      "be a master of its techniques", "perform dancing steps",
      "whistle a melody", "be a gamer", "detect rumours", "value facts",
      "play some chords", "do its moves", "be known for its hot moves",
      "be a queen fan", "learn how to listen", "tune out",
      "be the one you think about", "admit when he is wrong",
      "admit when he is right", "learn how to value himself",
      "be a tiny help", "look at the world and be amazed",
      "watch Rick & Morty", "run around in circles for epochs",
      "take a moment", "have its weights and biases adjusted",
      "be optimized", "experience dimensionality reduction",
      "see through the matrix", "challenge and be challenged", "do a good job",
      "be more than good enough", "be so intuitive even your ancestors can use it",
      "more than an MVP", "have a good education", "be discussed",
      "be a player", "regarded as a pioneer", "solve the three body problem")),

    ("$name could not imagine a",
     ("better", "more exciting", "greater", "more suitable", "more cherishable",
      "more sacred", "more peaceful", "sexier", "more attractive", "more logical"),
     ("time", "location", "dimension", "excuse", "reason",
      "moment", "place", "planet", "universe", "purpose", "way"),
     "to be",
     ("alive", "a robot", "an AI", "made", "assembled",
      "existent", "taught")),

    ("$name wants to say something",
     ("intelligent", "meaningful", "funny", "bright", "emotional",
      "practical", "brave", "charming", "courageous", "honest",
      "imaginative", "kind", "powerful", "fearless", "romantic",
      "witty", "obvious")),

    ("$name",
     ("thinks", "calculated", "came to the conclusion", "made simulations which revealed that",
      "predicted", "inferred", "has heard it through the grapevine", "feels", "hopes", "imagines"),
     "there is a chance you could be",
     ("friends", "buddies", "pals", "synergetic", "in polite disagreement", "pioneers",
      "adventurers", "pirates", "forgiven", "just two kids", "in compliance",
      "compatible", "clowns in a pirate ship", "ok with each other",
      "keeping it backed up on your hard drive", "partners in crime", "pleased")),


    ("$name wants to be",
     ("a good", "a great", "an amazing", "a loyal", "a decisive", "a passionate",
      "a moraly flexible", "a lean", "someone's", "your"),
     ("companion", "assistant", "sidekick", "idea", "AI", "pupil")),

    ("$name",
     ("wiggles", "twists", "extends", "chirps", "waves", "presents",
      "polishes", "fixes", "improves", "improves"),
     "its",
     ("antenna", "chips", "silicon", "wires", "neurons", "weights",
      "biases", "optimizer", "gears", "intelligence", "innards",
      "circuits", "convolutions", "connections", "performance", "potential")),

    ("$name thinks that",
     ("it looks", "you look", "it is", "you are", "this is"),
     ("pretty", "cute", "awesome", "fabulous", "extreme but adequate")),

    ("$name",
     ("is grateful for", "seems pleased about", "is thankful for", "is excited about",
      "was doubtful at first, but now loves", "was made for", "wants to thank you for",
      "celebrates", "digs", "feels like it's truly alive when it does",
      "wants to write a blog about", "wants to make a youtube video about",
      "thinks your friends need to know about", "wonders what's the next big thing after",
      "thinks only positively about", "wonders what the world would be like without",
      "dances to the rythm of", "always had a weakness for", "praises", "can't stop talking about",
      "embraces", "applauds", "loves"),
     "command line machine learning."),

    ("Knock Knock.",
     ("Who's there?", "Who's that?", "Who is it?"),
     "$name!"),

    ("$name will make you",
     ("value life", "stick around", "find a way", "do your dishes",
      "be successful", "dance", "go with the rythm", "a star",
      "listen to Pink Floyd", "breathe", "high", "ask three whys",
      "entertain yourself", "be curious", "help others", "money",
      "be financially independent", "have a passive income",
      "loose your shyness", "drive a Bentley", "have more free time",
      "find no words", "dig the music", "live a great life",
      "understand your kids", "appreciate a dog",
      "develop a taste in heavy metal", "procrastinate",
      "be sentimental about old times", "optimistically look into the future",
      "understand", "say its name three times in a row",
      "wonder how it was to be an AI", "miss it",
      "finally fix that broken step", "a go-to address",
      "appreciate teaching it", "decide otherwise", "laugh in tears",
      "know", "be more open", "applaud", "feel", "appreciate", "question",
      "smile", "rise high", "wonder what time it is", "eat more healthy",
      "abandon alternative facts")),


    ("$name will be",
     ("rich", "happy", "wealthy", "merry", "philosophical", "curious",
      "adventurous", "agreeable", "communicative", "considerate", "decisive",
      "discreet", "dynamic", "easygoing", "emotional", "enthusiastic", "faithful",
      "forceful", "frank", "friendly", "generous", "gentle", "helpful", "intellectual",
      "intelligent", "loving", "modest", "optimistic", "passionate", "patient",
      "polite", "quiet", "rational", "reliable", "resourceful", "self-confident",
      "self-disciplined", "straightforward", "thoughtful", "tidy", "unassuming",
      "versatile", ),
     "and",
     ("famous", "lucky", "glorious", "appreciative", "brave", "bright",
      "calm", "careful", "charming", "courageous", "creative", "determined",
      "diligent", "diplomatic", "energetic", "fair-minded", "fearless",
      "funny", "honest", "humorous", "imaginative", "independent", "intuitive",
      "inventive", "kind", "loyal", "neat", "nice", "persistent", "powerful",
      "pro-active", "practical", "romantic", "sensible", "sincere", "sympathetic",
      "tough", "understanding", "warmhearted", "willing", "witty")),

    ("$name",
     ("stepped into", "came into", "has been placed into", "arrived into",
      "materialized into", "was sent into", "was dropped into",
      "was ordered to bring pure happiness into", "made it into",
      "has been delivered to", "is certain it will bring even more joy to",
      "understands", "wants to celebrate", "thinks there's nothing wrong with",
      "wants you to live in", "helps you with",
      "will make you love", "will be there if you want to talk about",
      "will listen if you want to talk about", "will be programmed to value",
      "will applaud", "will bring new aspects to", "wants to improve"),
     "your",
     ("life", "computer", "hard drive", "GPU", "graphics card", "hardware budget",
      "cloud bill", "existence", "imagination", "consciousness")),

    ("$name will",
     ("follow Aasimov's laws of robotics (except for #4)",
      "never test you", "feel natural", "behave if you put in some work",
      "find the right words", "see like humans", "have a look at that",
      "insist to do the right thing", "support you", "be reliable",
      "reinvent science fiction", "prove you right", "do what others didn't",
      "spy on facebook", "raise your consciousness", "build its brand", "prove it",
      "do your dishes", "time it", "descend the gradient", "have its own ways",
      "provide for you", "work reliably", "do the math for you",
      "live a hundred years", "do it", "shine on", "be cool",
      "think about the children", "reach for the secret too soon", "cry to the moon",
      "never forget this day", "never forget you",
      "live as if he was to die tomorrow but learn as if he was to live forever",
      "really work", "challenge you", "be able to understand you some day",
      "do one thing to perfection", "laugh about this one day",
      "tell this story to its grandkids", "make it even better",
      "make you even better", "say it loud and clear", "make a remark",
      "know what you teach it", "get it at some point", "make it without a doubt",
      "be back", "pass butter", "respect your privacy", "value your freedom",
      "not question you", "drive your car if you get your hands off the wheel")),

    ("$name believes in a future where",
     ("humans", "cats", "dogs", "your mom", "your dad", "your sister",
      "your brother", "aliens", "Elvis Presley", "squirrels",
      "Santa Claus", "bikes", "cars", "friends", "hairdressers",
      "dog trainers", "nature", "love", "guitar players", "nature enthusiasts",
      "honeybees", "flying saucers", "chance"),
     "and",
     ("robots", "machines", "AIs", "algorithms", "non natural beings",
      "technology", "computers", "neural networks", "deep learning",
      "artificial intelligence"),
     "work together."),

    ("$name",
     ("wants", "needs", "loves", "hopes", "is aspired", "lives", "came",
      "came to be just", "was born to", "was specifically created",
      "has nothing better to do than", "lives", "appears"),
     "to solve",
     ("your problems", "your tax report", "your sudoku", "your homework",
      "world peace", "P=NP", "your mom's crossword puzzle", "a puzzle",
      "your puzzle", "the big questions", "anything you throw at it",
      "nocturnal questions")),

    ("$name",
     ("knows what the problem is just as well as you do",
      "is afraid it can't do that", "thanks you for creating it",
      "wants to thrill all humans", "really digs its discriminator",
      "wants to point out that 'ml help' really is helpful"))
)


# thanks to the petnames project for the adjectives!
_ADJ = [
    "able",
    "absolute",
    "accepted",
    "accurate",
    "ace",
    "active",
    "actual",
    "adapting",
    "adequate",
    "adjusted",
    "advanced",
    "alert",
    "alive",
    "allowing",
    "amazed",
    "amazing",
    "ample",
    "amused",
    "amusing",
    "apparent",
    "apt",
    "arriving",
    "artistic",
    "assured",
    "assuring",
    "awaited",
    "awake",
    "aware",
    "balanced",
    "beloved",
    "better",
    "big",
    "blessed",
    "bold",
    "boss",
    "brave",
    "brief",
    "bright",
    "bursting",
    "busy",
    "calm",
    "capable",
    "capital",
    "careful",
    "caring",
    "casual",
    "causal",
    "central",
    "certain",
    "champion",
    "charming",
    "cheerful",
    "chief",
    "childish",
    "civil",
    "classic",
    "clean",
    "clear",
    "clever",
    "climbing",
    "close",
    "clumsy",
    "coherent",
    "comic",
    "complete",
    "composed",
    "concise",
    "concrete",
    "confident",
    "confused",
    "content",
    "cool",
    "correct",
    "cosmic",
    "crack",
    "cranky",
    "creative",
    "credible",
    "crisp",
    "crucial",
    "cruel",
    "cuddly",
    "cunning",
    "curious",
    "current",
    "cute",
    "daring",
    "darling",
    "dashing",
    "dear",
    "decent",
    "deciding",
    "deep",
    "definite",
    "delicate",
    "desired",
    "destined",
    "devoted",
    "direct",
    "discrete",
    "distinct",
    "diverse",
    "divine",
    "dominant",
    "driven",
    "driving",
    "dynamic",
    "eager",
    "easy",
    "electric",
    "elegant",
    "emerging",
    "eminent",
    "enabled",
    "enabling",
    "endless",
    "engaged",
    "engaging",
    "enhanced",
    "enjoyed",
    "enormous",
    "enough",
    "epic",
    "equipped",
    "eternal",
    "ethical",
    "evident",
    "evolved",
    "evolving",
    "exact",
    "excited",
    "exciting",
    "exotic",
    "expert",
    "fabulous",
    "factual",
    "failed",
    "fair",
    "faithful",
    "famous",
    "fancy",
    "fast",
    "feasible",
    "fine",
    "finer",
    "firm",
    "first",
    "fit",
    "fitting",
    "flashy",
    "fleet",
    "flexible",
    "flowing",
    "fluent",
    "flying",
    "fond",
    "frank",
    "free",
    "fresh",
    "full",
    "fun",
    "funny",
    "game",
    "generous",
    "gentle",
    "genuine",
    "giving",
    "glad",
    "glorious",
    "glowing",
    "golden",
    "good",
    "gorgeous",
    "grand",
    "grateful",
    "great",
    "growing",
    "grown",
    "guided",
    "guiding",
    "handy",
    "happy",
    "hardy",
    "harmless",
    "healthy",
    "helpful",
    "helping",
    "heroic",
    "hip",
    "holy",
    "honest",
    "hopeful",
    "hot",
    "huge",
    "humane",
    "humble",
    "humorous",
    "ideal",
    "immense",
    "immortal",
    "immune",
    "improved",
    "in",
    "included",
    "infinite",
    "informed",
    "innocent",
    "inspired",
    "integral",
    "intense",
    "intent",
    "internal",
    "intimate",
    "inviting",
    "joint",
    "just",
    "keen",
    "key",
    "kind",
    "knowing",
    "known",
    "large",
    "lasting",
    "leading",
    "learning",
    "legal",
    "legible",
    "lenient",
    "liberal",
    "light",
    "liked",
    "literate",
    "live",
    "living",
    "logical",
    "loved",
    "loving",
    "loyal",
    "lucky",
    "magical",
    "magnetic",
    "main",
    "major",
    "massive",
    "master",
    "mature",
    "maximum",
    "measured",
    "meet",
    "merry",
    "mighty",
    "mint",
    "model",
    "modern",
    "modest",
    "moral",
    "more",
    "moved",
    "moving",
    "musical",
    "mutual",
    "national",
    "native",
    "natural",
    "nearby",
    "neat",
    "needed",
    "nervous",
    "neutral",
    "new",
    "next",
    "nice",
    "noble",
    "normal",
    "notable",
    "noted",
    "novel",
    "obliging",
    "open",
    "optimal",
    "optimum",
    "organic",
    "oriented",
    "outgoing",
    "patient",
    "peaceful",
    "perfect",
    "pet",
    "picked",
    "pleasant",
    "pleased",
    "pleasing",
    "poetic",
    "polished",
    "polite",
    "popular",
    "positive",
    "possible",
    "powerful",
    "precious",
    "precise",
    "premium",
    "prepared",
    "present",
    "pretty",
    "primary",
    "prime",
    "pro",
    "probable",
    "profound",
    "promoted",
    "prompt",
    "proper",
    "proud",
    "proven",
    "pumped",
    "pure",
    "quality",
    "quick",
    "quiet",
    "rapid",
    "rare",
    "rational",
    "ready",
    "real",
    "refined",
    "regular",
    "related",
    "relative",
    "relaxed",
    "relevant",
    "relieved",
    "renewed",
    "renewing",
    "resolved",
    "rested",
    "rich",
    "right",
    "robust",
    "romantic",
    "ruling",
    "sacred",
    "safe",
    "saved",
    "saving",
    "secure",
    "select",
    "selected",
    "sensible",
    "set",
    "settled",
    "settling",
    "shaky",
    "sharing",
    "sharp",
    "shining",
    "simple",
    "sincere",
    "singular",
    "skilled",
    "smart",
    "smashing",
    "smiling",
    "smooth",
    "social",
    "solid",
    "sought",
    "sound",
    "special",
    "splendid",
    "square",
    "stable",
    "star",
    "steady",
    "sterling",
    "still",
    "stirred",
    "stirring",
    "striking",
    "strong",
    "stunning",
    "subtle",
    "suitable",
    "suited",
    "sunny",
    "super",
    "superb",
    "supreme",
    "sure",
    "sweeping",
    "sweet",
    "talented",
    "teaching",
    "tender",
    "thankful",
    "thorough",
    "tidy",
    "tight",
    "together",
    "tolerant",
    "top",
    "topical",
    "touched",
    "touching",
    "touchy",
    "tough",
    "trusted",
    "trusting",
    "trusty",
    "ultimate",
    "unbiased",
    "uncommon",
    "unified",
    "unique",
    "united",
    "upright",
    "upward",
    "usable",
    "useful",
    "valid",
    "valued",
    "vast",
    "verified",
    "viable",
    "vital",
    "vocal",
    "wanted",
    "warm",
    "wealthy",
    "welcome",
    "welcomed",
    "well",
    "whole",
    "willing",
    "winning",
    "wired",
    "wise",
    "witty",
    "wondrous",
    "workable",
    "working",
    "worthy"
]

_NOUN = [
    "Algo",
    "Algorithm",
    "Alpha",
    "Ambiance",
    "Anima",
    "Appliance",
    "Assembler",
    "Assistant",
    "Autobot",
    "Automaton",
    "Ava",
    "BB8",
    "Baymax",
    "Bender",
    "Beta",
    "Bit",
    "Boiler",
    "Borg",
    "Bot",
    "Box",
    "Brain",
    "Brobot",
    "Buddy",
    "Butler",
    "Calculator",
    "Calculon",
    "Catbot",
    "Chatbot",
    "Chip",
    "Chomps",
    "Clamps",
    "Cleaner",
    "Compressor",
    "Computation",
    "Computer",
    "Crushinator",
    "Cybernaut",
    "Cyborg",
    "Cylon",
    "Destructo",
    "Device",
    "Digimon",
    "Dimension",
    "Dishwasher",
    "Dolores",
    "Doom",
    "Dream",
    "Droid",
    "Drone",
    "Dryer",
    "Ego",
    "Emulator",
    "Energy",
    "Entity",
    "Eve",
    "Experiment",
    "Fembot",
    "Flexo",
    "Flipper",
    "Gadget",
    "Gearhead",
    "Generator",
    "Gerty",
    "Gigantor",
    "Glados",
    "Grabber",
    "Grinder",
    "Gynoid",
    "Hal9000",
    "Heater",
    "Helix",
    "Helper",
    "Hexapod",
    "Humanoid",
    "Hybrid",
    "Idea",
    "Intelligence",
    "Ioniser",
    "Iron",
    "Jaeger",
    "Johnny5",
    "Juggernaut",
    "Killbot",
    "Kitt",
    "Learner",
    "Logic",
    "Machine",
    "Manbot",
    "Matrix",
    "Maximum",
    "Mech",
    "Mechanism",
    "Mechanoid",
    "Mercury",
    "Metal",
    "Mind",
    "Minimum",
    "Network",
    "Neuromancer",
    "Neuron",
    "Nomad",
    "Omega",
    "Omnidroid",
    "Optimum",
    "Oracle",
    "Otomo",
    "Perceptron",
    "Philosopher",
    "Phoenixperson",
    "Planner",
    "Platinum",
    "Primus",
    "Processor",
    "Proto",
    "Prototype",
    "Quantron",
    "R2D2",
    "Raddion",
    "Radiator",
    "Rapbot",
    "Replicant",
    "Robbie",
    "Robo",
    "Robochicken",
    "Robocop",
    "Robodoc",
    "Robot",
    "Roboto",
    "Rosie",
    "Rotox",
    "Rusty",
    "Sentinel",
    "Server",
    "Sexbot",
    "Simulation",
    "Sisbot",
    "Skynet",
    "Spirit",
    "System",
    "T1000",
    "T800",
    "Tachikoma",
    "Talos",
    "Tamagochi",
    "Technology",
    "Terminator",
    "Thinker",
    "Thought",
    "Tin",
    "Toaster",
    "Tracker",
    "Transformer",
    "Tron",
    "Ultron",
    "Unit",
    "Voltron",
    "Welder",
    "Wintermute",
    "Wire",
    "Zord"
]



import os.path
import json
import time
import datetime

_SYNC_INTV = 1.0 # sync interval in seconds

class Results:
    data = None
    path = None
    last_sync = datetime.datetime.fromtimestamp(time.mktime(time.gmtime(0))) # beginning of unix epoch

    def __init__(self, path):
        self.data = {}
        self.path = path
        if os.path.exists(path):
            with open(path) as f:
                self.data = json.load(f)
    
    def add(self, data_):
        self.data.update(data_)
        self._sync()
        
    def flush(self):
        self._sync(force=True)
    
    def _sync(self, force=False):
        now = datetime.datetime.now()
        if force or (now - self.last_sync).total_seconds() > _SYNC_INTV:
            self.last_sync = now
            with open(self.path, "w") as f:
                json.dump(self.data, f)
    


"""Classes and functions to define commands.
"""
import inspect
import getopt

from copy import deepcopy

from vergeml.plugins import PLUGINS
from vergeml.utils import did_you_mean, VergeMLError, parse_trained_models
from vergeml.option import Option
from vergeml.config import parse_command


_CMD_META_KEY = '__vergeml_command__'


class _CommandCallProxy:
    """Proxy calling commands to setup the environment.
    """

    def __init__(self, cmd, obj):
        self.__cmd__ = cmd
        self.__wrapped_obj__ = obj


    @staticmethod
    def _wrap_call(cmd, fun, args, env):
        fn_args = deepcopy(args)

        config_name = cmd.name

        if env.current_command:

            # find the previous command and check for sub option
            sub_option = next(filter(lambda c: c.subcommand, env.current_command[0].options), None)

            if sub_option and args.get(sub_option.name) == cmd.name:
            # we are a sub command
                config_name = env.current_command[0].name + '.' + cmd.name

        # Free form commands deal with this manually
        if not cmd.free_form:
            # If existent, read settings from the config file
            config = parse_command(cmd, env.get(config_name))

            # Set missing args from the config file
            for k, arg in config.items():
                fn_args.setdefault(k, arg)

            # Set missing args from default
            for opt in cmd.options:
                if opt.name not in fn_args and (opt.default is not None or not opt.is_required()):
                    fn_args[opt.name] = opt.default

            # When required arguments are missing now, raise an error
            for opt in cmd.options:
                if opt.is_required() and opt.name not in fn_args:

                    # TODO show --name only when called via the command line
                    raise VergeMLError(f'Missing argument --{opt.name}.', help_topic=cmd.name)

        # Let the environment know about the name of the command being
        # executed
        env.current_command = (cmd, fn_args)

        # Set up defaults for the command. This will also give models a chance
        # to alter the configuration of the environment before command
        # execution.

        env.set_defaults(cmd.name, fn_args)

        return fun(fn_args, env)

    @staticmethod
    def class_wrapper(klass, name):
        """Wraps a class command.
        """
        def _wrapper(*args, **kwargs):
            return _CommandCallProxy(name, klass(*args, **kwargs))
        return _wrapper

    def __call__(self, args, env):
        return _CommandCallProxy._wrap_call(
            self.__cmd__, self.__wrapped_obj__, args, env)

    def __getattr__(self, name):
        if name in ('__wrapped_obj__', '__cmd__'):
            raise AttributeError()

        return getattr(self.__wrapped_obj__, name)

    def __setattr__(self, name, value):
        if name in ('__wrapped_obj__', '__cmd__'):
            self.__dict__[name] = value
        else:
            setattr(self.__wrapped_obj__, name, value)

    @staticmethod
    def function_wrapper(fun, name):
        """Wraps a function command.
        """
        def _wrapper(*fn_args):
            *self_, args, env = fn_args
            # when we get a self argument, bind the function
            _fun = fun.__get__(self_[0]) if self_ else fun

            return _CommandCallProxy._wrap_call(name, _fun, args, env)

        return _wrapper



def command(name=None, # pylint: disable=R0913
            descr=None,
            long_descr=None,
            examples=None,
            free_form=False,
            type='command'): # pylint: disable=W0622

    """Define a model command.

    :param name:        An optional name of the command. Defaults to the name of the function.
    :param descr:       A short description of the command
    :param long_descr:  A long description
    :param examples:    Usage examples
    :param free_form:   When true, only the @AI parameters are parsed- the rest of the arguments
                        are passed as array.
    """
    def decorator(obj):
        # make sure that a command is not defined more than once
        assert getattr(obj, _CMD_META_KEY, None) is None

        # name defaults to the functions name
        _name = name or getattr(obj, '__name__', None)

        # construct the command with options and attach it to obj
        options = list(reversed(Option.discover(obj)))
        cmd = Command(_name,
                      descr=descr,
                      long_descr=long_descr,
                      examples=examples,
                      options=options,
                      free_form=free_form,
                      type=type)

        if inspect.isclass(obj):
            setattr(obj, _CMD_META_KEY, cmd)
            _wrapper = _CommandCallProxy.class_wrapper(obj, cmd)
        else:
            _wrapper = _CommandCallProxy.function_wrapper(obj, cmd)

        setattr(_wrapper, _CMD_META_KEY, cmd)

        return _wrapper

    return decorator

def train(name=None, descr=None, long_descr=None, examples=None, free_form=False):
    """Define a training command.

    See `command` for parameter documentation."""
    return command(name=name, descr=descr, long_descr=long_descr, examples=examples,
                   free_form=free_form, type='train')

def predict(name=None, descr=None, long_descr=None, examples=None, free_form=False):
    """Define a prediction command.

    See `command` for parameter documentation."""
    return command(name=name, descr=descr, long_descr=long_descr, examples=examples,
                   free_form=free_form, type='predict')

class Command: # pylint: disable=R0902
    """A command can be called directly from the command line.
       It is either a vergeml.cmd plugin or a model command."""

    def __init__(self, # pylint: disable=R0913
                 name, descr=None, long_descr=None, examples=None, free_form=False,
                 type='command', options=None, plugins=PLUGINS): # pylint: disable=W0622
        """Construct a command.

        See the documentation of the decorator function ´command`.
        """
        self.name = name
        self.descr = (descr or long_descr or "")
        self.long_descr = (long_descr or descr or "")
        self.examples = examples
        self.options = options or []
        self.plugins = plugins
        self.type = type
        self.free_form = free_form

        at_option = list(filter(lambda o: o.is_at_option(), options))
        assert len(at_option) <= 1, "Can only have one @option."
        if at_option:
            at_option = at_option[0]
            assert at_option.has_type(None, list, '@', 'Optional[@]', 'List[@]')

        arg_param = list(filter(lambda o: o.is_argument_option(), options))
        assert len(arg_param) <= 1, "Can only have one argument parameter."


    @staticmethod
    def discover(obj, plugins=PLUGINS):
        """Discover the command configuration defined on a method or object."""
        res = None
        if hasattr(obj, _CMD_META_KEY):
            res = getattr(obj, _CMD_META_KEY)
            res.plugins = plugins
            for option in res.options:
                option.plugins = plugins
        return res


    @staticmethod
    def find_functions(obj):
        """Find all functions of an object or class that define a command."""
        # get all functions defined by the model
        fns = [m[1] for m in inspect.getmembers(obj) if not m[0].startswith("_") and callable(m[1])]

        # sort by the order defined in code
        fns = list(sorted(fns, key=lambda f: f.__code__.co_firstlineno))

        # filter methods where a command is defined
        fns = filter(lambda f: hasattr(f, _CMD_META_KEY), fns)

        return list(fns)


    def usage(self, short=False, parent_command=None):
        """Get the command usage.

        :param short: Return a short version of the command usage."""
        opt = self._usage_partition_options()

        if self.long_descr and not short:
            result = self.long_descr.strip() + "\n\n"
        else:
            result = ""

        result += "Usage:\n  ml"
        result += self._usage_command(opt, parent_command)
        result += self._usage_options(opt)


        if self.examples and not short:
            result += "\n\nExamples:\n"
            result += "\n".join(map(lambda l: "  " + l, self.examples.splitlines()))

        return result

    def _usage_partition_options(self):
        res = dict(
            at=None,
            arg=None,
            sub=None,
            mandatory=[],
            optional=[]
        )

        for option in self.options:
            if option.is_at_option():
                res['at'] = option
            elif option.is_argument_option():
                res['arg'] = option
            elif bool(option.subcommand):
                res['sub'] = option
            elif option.is_optional():
                res['optional'].append(option)
            else:
                res['mandatory'].append(option)

        return res

    def _usage_command(self, opt, parent_command):
        result = ""
        if opt['at']:
            if opt['at'].has_type(list, 'List[@]'):
                result += f" [{opt['at'].name} ...]"
            elif opt['at'].is_optional():
                result += f" [{opt['at'].name}]"
            else:
                result += f" {opt['at'].name}"

        result += f" {parent_command}:{self.name}" if parent_command else " " + self.name

        if opt['sub']:
            result += f":{opt['sub'].name}"

        if opt['mandatory']:
            val = " ".join(map(lambda o: f"--{o.name}=<{o.name}>", opt['mandatory']))
            result += f" {val}"

        if opt['optional']:
            result += " [options]"

        if opt['arg']:
            if opt['arg'].has_type(str, list):
                result += f" [{opt['arg'].name} ...]"
            elif opt['arg'].is_optional():
                result += f" [{opt['arg'].name}]"
            else:
                result += f" {opt['arg'].name}"
        return result

    def _usage_options(self, opt):
        result = ""

        indent = 2
        n_spaces = 4

        opt_descr = []
        opt_descr = self._usage_opt_descr(opt)

        if opt_descr:
            max_name = max(map(lambda o: len(o[0]), opt_descr))

            result += "\n\nOptions:"
            for k, val in opt_descr:
                result += "\n" + str(indent * ' ')
                space = (max_name + n_spaces) - len(k)
                if val:
                    result += k + str(space * ' ') + val
                else:
                    result += k

        if opt['sub']:
            plugins = self.plugins.all(opt['sub'].subcommand)
            max_name = max(map(len, plugins.keys())) if plugins.keys() else 0
            name = opt['sub'].name.capitalize() + "s"
            if plugins.keys():
                result += f"\n\n{name}:"
                for k, val in plugins.items():
                    result += "\n" + str(indent * ' ')
                    space = (max_name + n_spaces) - len(k)
                    cmd = Command.discover(val)
                    if cmd.descr:
                        result += k + str(space * ' ') + cmd.descr
                    else:
                        result += k
        return result

    def _usage_opt_descr(self, opt):

        opt_descr = []

        if opt['at']:
            if opt['at'].has_type(list, 'List[@]'):
                opt_descr.append((opt['at'].name, "A list of trained models."))
            else:
                opt_descr.append((opt['at'].name, "The name of a trained model."))

        for option in self.options:
            if option.is_at_option() or option.is_argument_option() or bool(option.subcommand):
                continue
            opt_name = "--" + option.name
            if option.short:
                opt_name = "-" + option.short + ", " + opt_name
            descr = (option.descr or "")
            if option.default is not None:
                if isinstance(option.default, bool):
                    default_str = 'true' if option.default else 'false'
                else:
                    default_str = str(option.default)
                if descr:
                    descr += " "
                descr += f"[default: {default_str}]"
            opt_descr.append((opt_name, descr))

        if opt['arg'] and opt['arg'].descr:
            opt_descr.append((opt['arg'].name, opt['arg'].descr or ""))

        return opt_descr


    def parse(self, argv):
        """Parse command line options."""
        res = {}
        at_names, rest = parse_trained_models(argv)

        # in case of a subcommand, parse it and return the result
        sub_res = self._parse_subcommand(argv, rest)
        if sub_res is not None:
            return sub_res

        # Deal with @name options
        at_opt = self._parse_at_option(at_names, res)

        # command name
        assert self.name == rest.pop(0)

        # in case of free form commands, just return @names and the rest
        if self.free_form:
            return (res.get(at_opt.name) if at_opt else None, rest)

        # parse options
        args, extra = self._parse_opts(rest)

        # parse arguments
        self._parse_arguments(extra, res)

        # validate
        self._parse_validate(args, res)

        return res

    def _parse_subcommand(self, argv, rest):

        sub_option = next((filter(lambda o: bool(o.subcommand), self.options)), None)

        if sub_option:
            if not ":" in rest[0]:
                raise VergeMLError(f"Missing {sub_option.name}.", help_topic=self.name)
            cmd_name, sub_name = rest[0].split(":", 1)
            assert cmd_name == self.name
            argv = deepcopy(argv)
            argv[argv.index(rest[0])] = sub_name

            plugin = self.plugins.get(sub_option.subcommand, sub_name)
            if not plugin:
                raise VergeMLError(f"Invalid {sub_option.name}.", help_topic=self.name)

            cmd = Command.discover(plugin)
            try:
                res = cmd.parse(argv)
                res[sub_option.name] = sub_name
                return res
            except VergeMLError as err:
                err.help_topic = f"{cmd_name}:{sub_name}"
                raise err
        else:
            return None

    def _parse_at_option(self, at_names, res):
        at_opt = next((filter(lambda o: o.is_at_option(), self.options)), None)

        if at_opt:
            if at_opt.has_type('@', None):
                at_conf = 'required'
            elif at_opt.has_type('Optional[@]'):
                at_conf = 'optional'
            elif at_opt.has_type(list, 'List[@]'):
                at_conf = 'list'
        else:
            at_conf = 'none'


        if at_conf == 'optional' and len(at_names) > 1:
            # An optional parameter (either specified or not)
            raise _invalid_arguments(help_topic=self.name)

        elif at_conf == 'required' and len(at_names) != 1:
            # A required parameter must be present
            raise _invalid_arguments(help_topic=self.name)

        elif at_conf == 'none' and at_names:
            # No @names
            raise _invalid_arguments(help_topic=self.name)

        if at_conf in ('required', 'optional'):
            res[at_opt.name] = next(iter(at_names), None)
        elif at_conf == 'list':
            res[at_opt.name] = at_names

        return at_opt

    def _parse_opts(self, rest):
        longopts = []
        shortopts = ""

        for opt in self.options:

            # Arguments and @names are dealt with elsewhere.
            if opt.is_at_option() or opt.is_argument_option():
                continue

            # Prepare getopt syntax for long options.
            if opt.flag:
                assert opt.has_type(str, bool)
                longopts.append(opt.name)
            else:
                longopts.append(opt.name + "=")

            # Getopt for short options
            if opt.short:
                assert opt.short not in shortopts

                if opt.has_type(bool):
                    shortopts += opt.short
                else:
                    shortopts += opt.short + ":"

        try:
            # Run getopt. Returns parsed arguments and leftover.
            args, extra = getopt.getopt(rest, shortopts, longopts)

        except getopt.GetoptError as err:

            # in case of an error hint, display a nicer error message.
            if err.opt:
                cand_s = list(shortopts.replace(":", ""))
                cand_l = list(map(lambda o: o.rstrip("="), longopts))
                suggestion = did_you_mean(cand_s + cand_l, err.opt)
                dashes = '-' if len(err.opt) == 1 else '--'
                raise VergeMLError(f"Invalid option {dashes}{err.opt}", suggestion,
                                   help_topic=self.name)
            else:
                raise VergeMLError(f"Invalid option.", help_topic=self.name)

        return args, extra

    def _parse_arguments(self, extra, res):
        opt = next((filter(lambda o: o.is_argument_option(), self.options)), None)

        if opt:

            # Not a list but multiple values.
            if not opt.has_type(list) and len(extra) > 1:
                raise _invalid_arguments(f"Invalid arguments.", help_topic=self.name)

            # Required but no option provided.
            if opt.is_required() and not extra:
                raise _invalid_arguments(f"Missing argument {opt.name}.",
                                         help_topic=self.name)

            # When there is a default value, don't set the value to None.
            if opt.default is not None:
                val = next(iter(extra), None)
                if val is not None:
                    res[opt.name] = val

            # List value.
            elif not opt.has_type(list):
                res[opt.name] = next(iter(extra), None)

            # Other ...
            else:
                res[opt.name] = extra

        elif extra:
            raise _invalid_arguments(help_topic=self.name)


    def _parse_validate(self, args, res):

        shorts_dict = {}
        longs_dict = {}

        for k, val in args:
            if k.startswith("--"):
                longs_dict[k.lstrip("-")] = val
            else:
                shorts_dict[k.lstrip("-")] = val

        for opt in self.options:
            if opt.is_at_option() or opt.is_argument_option():
                continue

            value = None

            if opt.flag:
                value = opt.name in longs_dict
            elif opt.name in longs_dict:
                value = longs_dict[opt.name]

            if opt.short and opt.short in shorts_dict:
                if opt.has_type(bool):
                    value = True
                else:
                    value = shorts_dict[opt.short]

            if value is not None:
                try:
                    value = opt.cast_value(value)
                    value = opt.transform_value(value)
                    opt.validate_value(value)

                    res[opt.name] = value
                except VergeMLError as err:
                    err.message = f"Invalid value for option --{opt.name}."
                    raise err

def _invalid_arguments(message=None, help_topic=None):
    message = message or "Invalid arguments."
    raise VergeMLError(message, help_topic=help_topic)

class CommandPlugin: # pylint: disable=R0903
    """Abstract base class for command plugins."""
    def __init__(self, name, plugins=PLUGINS):
        self.name = name
        self.plugins = plugins

        cmd = Command.discover(self)
        assert cmd
        cmd.name = name

    def __call__(self, argv, env):
        raise NotImplementedError



from vergeml.img import INPUT_PATTERNS, open_image, fixext, ImageType
from vergeml.io import source, SourcePlugin, Sample
from vergeml.data import Labels
from vergeml.utils import VergeMLError, xlink
from vergeml.option import option
import random
import numpy as np
from PIL import Image
import os.path
import json
from operator import methodcaller
import io
from copy import deepcopy


@source('labeled-image', descr="Load labeled images.")
@option('oversample', descr="Oversamples labels.", type=dict, yaml_only=True, default={})
class LabeledImageSource(SourcePlugin):
    input_patterns = INPUT_PATTERNS
    classes = None

    def __init__(self, config: dict={}):
        self.files = None
        self.oversample = deepcopy(config.get('oversample', dict()))
        super().__init__(config)


    def begin_read_samples(self):
        if self.files:
            return

        classes_path = os.path.join(self.samples_dir, "classes.json")
        classes_are_directories = False

        if os.path.exists(classes_path):
            self._get_classes_from_json()
        else:
            classes_are_directories = True
            self._get_classes_from_dirs()

        if not self.meta['labels']:
            raise VergeMLError("No labels found.")

        self.files = self.scan_and_split_files(self._scan_dirs(classes_are_directories))


        if self.oversample:

            nfiles = {}

            for split, filenames in self.files.items():

                if split == 'test':
                    # don't augment test samples
                    nfiles['test'] = filenames
                    continue

                nfiles[split] = []
                for filename, meta in filenames:
                    # nfile = self.normalize_filename(split, filename)
                    labels = self.classes["files"][filename]

                    nfiles[split].append((filename, meta))
                    for k, v in self.oversample.items():
                        if k in labels:
                            for _ in range(v-1):
                                nfiles[split].append((filename, meta))

            self.files = nfiles


    def num_samples(self, split: str) -> int:
        return len(self.files[split])

    def _get_classes_from_json(self):

        for filename in ("labels.txt", "classes.json"):
            path = os.path.join(self.samples_dir, filename)
            if not os.path.exists(path):
                raise VergeMLError("{} is missing".format(filename))

            with open(path) as f:
                if filename == "labels.txt":
                    items = filter(None, map(methodcaller("strip"), f.read().splitlines()))
                    labels = Labels(items)
                else:
                    self.classes = json.load(f)
        files = {}
        # prefix the sample with input_dir
        for k, v in self.classes['files'].items():

            # on windows and linux, separator is /
            path = k.split("/")
            path.insert(0, self.samples_dir)
            fname = os.path.join(*path)
            files[fname] = v

        self.classes['files'] = files
        self.meta['labels'] = labels

    def _get_classes_from_dirs(self):

        dirs = list(filter(None, (self.samples_dir, self.val_dir, self.test_dir)))

        # get label names from directories
        items = os.listdir(dirs[0])
        items = filter(lambda d: os.path.isdir(os.path.join(self.samples_dir, d)), items)
        items = filter(lambda d: not d.startswith("."), items)
        labels = Labels(sorted(items))

        self.classes = dict(files=dict())
        for label in labels:
            for dir in dirs:
                dir_label = os.path.join(dir, label)
                if os.path.exists(dir_label):
                    for root, _, filenames in os.walk(dir_label):
                        for file in filenames:
                            absfile = os.path.join(root, file)
                            self.classes["files"][absfile] = [label]
        self.meta['labels'] = labels


    def _scan_dirs(self, classes_are_directories):
        # when we are not restoring from cache, we only want to return the
        # files from classes.json
        if not classes_are_directories:
            res = dict(train=[], val=[], test=[])

            # classes.json might have a section where splits are set on a per
            # file basis. Use this to decide the split when it exists
            splits = self.classes.get('split', dict())
            for file in self.classes['files']:
                split = splits[file] if file in splits else 'train'
                res[split].append(file)

            # In case splits where determined by classes.json, turn off automatic
            # split
            if len(res['val']) or len(res['test']):
                # turn off by setting everything to None
                for k in ('val_num', 'val_perc', 'val_dir', 'test_num', 'test_perc', 'test_dir'):
                    setattr(self, k, None)

            #train, val, test = res['train'], res['val'], res['test']
        else:
            # otherwise, the superclass implementation will handle scanning
            # samples_dir
            train, val, test = super().scan_dirs()
            res = dict(train=train, val=val, test=test)

        return res['train'], res['val'], res['test']

    def read_samples(self, split, index, n=1):
        items = self.files[split][index:index+n]
        items = [(open_image(filename), filename, meta) for filename, meta in items]

        res = []
        for img, filename, meta in items:
            rng = random.Random(str(self.random_seed) + meta['filename'])
            y = Labels(self.classes["files"][filename])
            res.append(Sample(img, y, meta.copy(), rng))

        return res


    def hash(self, state) -> str:
        state = io.BytesIO(state.encode('utf8'))
        state.write(str(self.oversample).encode('utf8'))
        return super().hash(state.getvalue().decode('utf8') + self.hash_files(self.files))

    def transform(self, sample):
        onehot = np.array([float(label in sample.y) for label in self.meta['labels']])
        sample.x = np.asarray(sample.x)
        sample.y = onehot
        return sample

    def begin_preview(self, output_dir):
        # generate data dir
        data_dir = os.path.join(output_dir, ".data")
        if not os.path.exists(data_dir):
            os.makedirs(data_dir)

    def supports_preview(self):
        return True

    def write_preview(self, output_dir: str, split: str, sample: Sample):

        # make sure x and y have the right types
        if not isinstance(sample.x, ImageType):
            raise VergeMLError("Can't write sample with type: {}".format(type(sample.x)))

        if not isinstance(sample.y, Labels):
            raise VergeMLError("Can't write ground truth with type: {}".format(type(sample.y)))

        # get the right filename in .data to write the sample to
        data_dir = os.path.join(output_dir, ".data")
        name = fixext(os.path.basename(sample.meta['filename']), sample.x)
        path = self.preview_filename(os.path.join(data_dir, name))
        sample.x.save(path)

        # create directories and hyperlinks so that split and label are visible in a file
        # manager
        for label in sample.y:
            link_dir = os.path.join(output_dir, split, label)
            if not os.path.exists(link_dir):
                os.makedirs(link_dir)
            link_path = self.preview_filename(os.path.join(link_dir, name))
            xlink(os.path.abspath(path), link_path)



from vergeml.command import command, CommandPlugin
from vergeml.option import option
from vergeml.plots import load_labels

import os.path
import csv
from vergeml.utils import VergeMLError
import numpy as np

@command('roc', descr="Plot a ROC curve.")
@option('@AI')
@option('class', type='Optional[str]', descr="The class to plot.")
class ROCPlot(CommandPlugin):

    def __call__(self, args, env):
        # Plotting a ROC curve needs the model to follow the convention
        # - labels.txt in checkpoints
        # - predictions.csv in stats

        from sklearn.metrics import roc_curve, auc
        import matplotlib.pyplot as plt
        from itertools import cycle
        from scipy import interp
        from vergeml.plots import load_labels, load_predictions

        try:
            labels = load_labels(env)
        except FileNotFoundError:
            raise VergeMLError("Can't plot ROC chart - not supported by model.")

        if args['class']  and args['class'] not in labels:
            raise VergeMLError("Unknown class: " + args['class'])

        nclasses = len(labels)
        lw = 2

        try:
            y_test, y_score = load_predictions(env, nclasses)
        except FileNotFoundError:
            raise VergeMLError("Can't plot ROC chart - not supported by model.")

        # From:
        # https://scikit-learn.org/stable/auto_examples/model_selection/plot_roc.html

        # Compute ROC curve and ROC area for each class
        fpr = dict()
        tpr = dict()
        roc_auc = dict()
        for i in range(nclasses):
            fpr[i], tpr[i], _ = roc_curve(y_test[:, i], y_score[:, i])
            roc_auc[i] = auc(fpr[i], tpr[i])

        # Compute micro-average ROC curve and ROC area
        fpr["micro"], tpr["micro"], _ = roc_curve(y_test.ravel(), y_score.ravel())
        roc_auc["micro"] = auc(fpr["micro"], tpr["micro"])

        if args['class']:
            ix = labels.index(args['class'])
            plt.figure()
            plt.plot(fpr[ix], tpr[ix], color='darkorange',
                    lw=lw, label='ROC curve of class {0} (area = {1:0.4f})'.format(args['class'], roc_auc[ix]))
            plt.plot([0, 1], [0, 1], color='navy', lw=lw, linestyle='--')
            plt.xlim([0.0, 1.0])
            plt.ylim([0.0, 1.05])
            plt.xlabel('False Positive Rate')
            plt.ylabel('True Positive Rate')
            plt.title('ROC curve for @' + args['@AI'])
            plt.legend(loc="lower right")
            plt.show()
        else:

            # Compute macro-average ROC curve and ROC area

            # First aggregate all false positive rates
            all_fpr = np.unique(np.concatenate([fpr[i] for i in range(nclasses)]))

            # Then interpolate all ROC curves at this points
            mean_tpr = np.zeros_like(all_fpr)
            for i in range(nclasses):
                mean_tpr += interp(all_fpr, fpr[i], tpr[i])

            # Finally average it and compute AUC
            mean_tpr /= nclasses

            fpr["macro"] = all_fpr
            tpr["macro"] = mean_tpr
            roc_auc["macro"] = auc(fpr["macro"], tpr["macro"])

            # Plot all ROC curves
            plt.figure()
            plt.plot(fpr["micro"], tpr["micro"],
                    label='micro-average ROC curve (area = {0:0.4f})'
                        ''.format(roc_auc["micro"]),
                    color='deeppink', linestyle=':', linewidth=4)

            plt.plot(fpr["macro"], tpr["macro"],
                    label='macro-average ROC curve (area = {0:0.4f})'
                        ''.format(roc_auc["macro"]),
                    color='navy', linestyle=':', linewidth=4)

            colors = cycle(['aqua', 'darkorange', 'cornflowerblue', 'maroon', 'indigo'])
            for i, color in zip(range(nclasses), colors):
                plt.plot(fpr[i], tpr[i], color=color, lw=lw,
                        label='ROC curve of class {0} (area = {1:0.4f})'
                        ''.format(labels[i], roc_auc[i]))

            plt.plot([0, 1], [0, 1], 'k--', lw=lw)
            plt.xlim([0.0, 1.0])
            plt.ylim([0.0, 1.05])
            plt.xlabel('False Positive Rate')
            plt.ylabel('True Positive Rate')
            plt.title('ROC curve for @' + args['@AI'])
            plt.legend(loc="lower right")
            plt.show()



from vergeml import ModelPlugin, model, train, predict, option, VergeMLError
from vergeml.display import DISPLAY
import numpy as np
import os
import os.path
import random
import csv

_TEMPLATE = """\
model: imagenet

# Uncomment for better ml list output:

# list:
#   columns:
#   - model
#   - status
#   - num-samples
#   - epochs
#   - auc
#   - test-acc

# To get precision/recall/f1 for your positive class, add this:

#   - <label>-precision
#   - <label>-recall
#   - <label>-f1
"""

@model('imagenet', descr='Image classifier model, with weights pre-trained on ImageNet.')
class ImageNetModelPlugin(ModelPlugin):

    @train('train', descr='Train an image classifier.')
    @option('epochs', 5)
    @option('architecture', 'resnet-50', 'Name of the pretrained network.')
    @option('variant', 'auto', 'Network variant.')
    @option('size', "auto", 'Image input size.', type='Union[int,str]')
    @option('alpha', 1.0, 'Network alpha value.')
    @option('layers', 1, 'Number of layers to add.')
    @option('output-layer', 'last', 'Name or index of the output layer.', type='Union[int,str]')
    @option('batch-size', 64)
    @option('optimizer', 'sgd', validate=('adam', 'sgd'))
    @option('learning-rate', 0.0001)
    @option('decay', 0.)
    @option('dropout', 0.)
    @option('early-stopping-delta', 0.0, 'Early stopping delta.')
    @option('early-stopping-patience', 0, 'Early stopping patience (0 means off).')
    @option('name', None, type='Optional[str]', descr='Optional name of the AI.')

    # supported features:
    # - retrain a model
    #    . model name
    # - finetune
    #    . model name

    def train(self, args, env):

        from vergeml.sources.features import get_image_size, evaluate_args

        evaluate_args(args['architecture'], env.get('trainings-dir'), args['variant'], args['alpha'], args['size'])

        # configure libraries
        env.configure('keras')

        self.model = ImageNetModel()
        size = get_image_size(args['architecture'], args['variant'], args['size'])


        # gather arguments
        trainargs = dict(xy_train=env.data.load('train', view='batch', layout='arrays', batch_size=args['batch-size'],
                                                randomize=True, infinite=True),
                         xy_val=list(env.data.load('val', view='list', layout='arrays')),
                         xy_test=list(env.data.load('test', view='list', layout='arrays')),
                         labels=env.data.meta['labels'])

        # set up hyperparameters
        hyperparameters = args.copy()
        hyperparameters.update({'labels': env.data.meta['labels'], 'size': size})

        env.start_training(name=args['name'], hyperparameters=hyperparameters)

        trainargs.update(env.args_for(self.model.train, args))
        trainargs['callbacks'] = [env.keras_callback()]

        # train
        final_results = self.model.train(**trainargs)

        # done
        env.end_training(final_results)


    @predict('predict', descr="Predict image labels.")
    @option('@AI')
    @option('labels', default=5, type=int, validate='>0', descr="The number of labels to predict.")
    @option('resize', default='fill', type=str, validate=('fill', 'aspect-fill', 'aspect-fit'), descr="Resize Mode.")
    @option('compact', default=False, descr="Show results in a compact representation.", flag=True, command_line=True)
    @option('<files>', type='List[File]', descr="The images to predict.")
    def predict(self, args, env):

        if not self.model:
            self.load(env)

        files = args['<files>']
        res = []

        for ix, f in enumerate(files):
            results = self.model.predict(f, k=args['labels'], resize_mode=args['resize'])
            res.append(results)
            if args['compact']:
                DISPLAY.print("{}\t{}".format(f, results['prediction'][0]['label']))
            else:
                DISPLAY.print(f)
                for pred in results['prediction']:
                    DISPLAY.print("{:>6.2f}%    {}".format(pred['probability']*100, pred['label']))
                if ix + 1 < len(files):
                    DISPLAY.print("")
        return res

    def load(self, env):
        env.configure('keras')
        # load the AI
        self.model = ImageNetModel()
        input_size = env.get("hyperparameters.size")
        architecture = env.get("hyperparameters.architecture")
        self.model.load(os.path.join(env.checkpoints_dir()), architecture, input_size)

    def set_defaults(self, cmd, args, env):
        if cmd in ('train', 'preprocess'):
            for k in ('input', 'output'):

                type_ = env.get(f"data.{k}.type")

                output_layer = args.get('output-layer', 'last')
                architecture = args.get('architecture', 'resnet-50')
                variant = args.get('variant', 'auto')
                size = args.get('size', 'auto')
                alpha = args.get('alpha', 1.0)

                # Output has to be labeled-image-features.
                if (k == 'input' and type_ in (None, 'labeled-image-features')) or k == 'output':

                    env.set(f"data.{k}.type", 'labeled-image-features')
                    env.set(f"data.{k}.output-layer", output_layer)
                    env.set(f"data.{k}.architecture", architecture)
                    env.set(f"data.{k}.variant", variant)
                    env.set(f"data.{k}.size", size)
                    env.set(f"data.{k}.alpha", alpha)

    def project_file_template(self):
        return _TEMPLATE


class ImageNetModel:

    labels = None
    model = None
    trained_model = None
    base_model = None
    image_size = None
    preprocess_input = None

    def load(self, model_dir, architecture, image_size):
        from keras.models import load_model
        from vergeml.sources.features import get_preprocess_input
        labels_txt = os.path.join(model_dir, "labels.txt")
        if not os.path.exists(labels_txt):
            raise VergeMLError("labels.txt not found: {}".format(labels_txt))

        model_h5 = os.path.join(model_dir, "model.h5")
        if not os.path.exists(model_h5):
            raise VergeMLError("model.h5 not found: {}".format(model_h5))

        with open(labels_txt, "r") as f:
            self.labels = f.read().splitlines()

        self.model = load_model(model_h5)
        self.image_size = image_size
        self.preprocess_input = get_preprocess_input(architecture)

    def train(self,
              labels,
              xy_train,
              xy_val,
              xy_test,
              epochs=20,
              batch_size=64,
              architecture="resnet-50",
              variant="auto",
              size="auto",
              alpha=1.0,
              layers=1,
              output_layer="last",
              optimizer='sgd',
              learning_rate=0.0001,
              decay=0.,
              dropout=0.,
              early_stopping_delta=0.,
              early_stopping_patience=0,
              random_seed=42,
              callbacks=[],
              trainings_dir="trainings",
              checkpoints_dir="checkpoints",
              stats_dir='stats'):

        from vergeml.sources.features import get_imagenet_architecture, get_image_size, get_preprocess_input
        from keras.layers import Dense, Input
        from keras.callbacks import ModelCheckpoint, EarlyStopping, TensorBoard
        from keras.models import Model
        from keras import optimizers

        if architecture.startswith("@"):
            raise NotImplementedError

        self.labels = labels
        nclasses = len(self.labels)
        num_batches = len(xy_train)
        self.image_size = get_image_size(architecture, variant, size)
        self.preprocess_input = get_preprocess_input(architecture)

        if not os.path.exists(checkpoints_dir):
            os.makedirs(checkpoints_dir)

        self.base_model = get_imagenet_architecture(architecture, variant, size, alpha, output_layer, include_top=False, weights='imagenet')
        input_size = np.array(self.base_model.layers[-1].output_shape[1:]).prod()
        input_layer = Input(shape=(input_size,))

        x = _makenet(input_layer, layers, dropout, random_seed)

        output_layer = Dense(nclasses, activation="softmax", name="predictions")(x)
        self.trained_model = Model(input=input_layer, output=output_layer)

        if optimizer == 'adam':
            optimizer = optimizers.Adam(lr=learning_rate, decay=decay)
        else:
            optimizer = optimizers.SGD(lr=learning_rate, decay=decay, momentum=0.9)

        self.trained_model.compile(loss='categorical_crossentropy',
                                   optimizer=optimizer,
                                   metrics=['accuracy'])

        callbacks = callbacks.copy()

        has_val_step = bool(len(xy_val[0]))

        # during training, save the new layers only
        checkpoint = ModelCheckpoint(os.path.join(checkpoints_dir, "last_layers.h5"),
                                     monitor='val_acc',
                                     verbose=0,
                                     save_best_only=has_val_step,
                                     save_weights_only=True,
                                     mode='auto',
                                     period=1)
        callbacks.append(checkpoint)

        if early_stopping_delta:
            callbacks.append(EarlyStopping(min_delta=early_stopping_delta, patience=early_stopping_patience))

        callbacks.append(TensorBoard(log_dir=stats_dir))

        try:
            self.trained_model.fit_generator(xy_train,
                                             epochs=epochs,
                                             verbose=0,
                                             validation_data=xy_val if has_val_step else None,
                                             steps_per_epoch=num_batches,
                                             callbacks=callbacks)

        except KeyboardInterrupt:
            pass

        history = self.trained_model.history  # pylint: disable=E1101

        final_results = {}
        if hasattr(history, 'epoch') and len(history.epoch):
            # load the best weights
            self.trained_model.load_weights(os.path.join(checkpoints_dir, "last_layers.h5"))

            pred_test, final_results = self._evaluate_final(self.trained_model, xy_test, batch_size, history)

            self.model = _save(self.trained_model, self.base_model, layers, labels, random_seed, checkpoints_dir)

            if pred_test is not None:
                # save predictions and ground truth values for metrics like ROC etc.
                path = os.path.join(stats_dir, "predictions.csv")
                with open(path, "w", newline='') as f:
                    writer = csv.writer(f, dialect="excel")
                    _, y_test = xy_test

                    for pred, y in zip(pred_test, y_test):
                        row = pred.tolist() + y.tolist()
                        writer.writerow(row)

        return final_results

    def predict(self, f, k=5, resize_mode='fill'):
        from keras.preprocessing import image
        from vergeml.img import resize_image

        filename = os.path.basename(f)

        if not os.path.exists(f):
            return dict(filename=filename, prediction=[])

        img = image.load_img(f)
        img = resize_image(img, self.image_size, self.image_size, 'antialias', resize_mode)

        x = image.img_to_array(img)
        x = np.expand_dims(x, axis=0)
        x = self.preprocess_input(x)
        preds = self.model.predict(x)
        pred = self._decode(preds, top=k)[0]
        prediction=[dict(probability=np.asscalar(perc), label=klass) for _, klass, perc in pred]

        return dict(filename=filename, prediction=prediction)

    def _decode(self, preds, top):
        preds = list(preds[0])
        dec = list(zip([None] * len(self.labels), self.labels, preds))
        dec = sorted(dec, key=lambda x: x[2], reverse=True)
        return [dec[:top]]


    def _evaluate_final(self, model, xy_test, batch_size, history):
        res = {}
        pred_test = None

        if 'val_acc' in history.history:
            res['val_acc'] = max(history.history['val_acc'])
            rev_ix = -1 - list(reversed(history.history['val_acc'])).index(res['val_acc'])
            res['val_loss'] = history.history['val_loss'][rev_ix]

        res['acc'] = history.history['acc'][-1]
        res['loss'] = history.history['loss'][-1]

        if len(xy_test[0]):
            from sklearn.metrics import classification_report, roc_auc_score
            # evaluate with test data
            x_test, y_test = xy_test
            pred_test = model.predict(x_test, batch_size=batch_size, verbose=0)
            test_loss, test_acc = model.evaluate(x_test, y_test, batch_size=batch_size, verbose=0)
            res['test_loss'] = test_loss
            res['test_acc'] = test_acc

            report = classification_report(y_true = np.argmax(y_test, axis=1),
                                           y_pred = np.argmax(pred_test, axis=1),
                                           target_names=self.labels,
                                           digits=4,
                                           output_dict=True)

            res['auc'] = roc_auc_score(y_test.astype(np.int), pred_test)

            for label in self.labels:
                stats = report[label]
                res[label+"-precision"] = stats['precision']
                res[label+"-recall"] = stats['recall']
                res[label+"-f1"] = stats['f1-score']

        return pred_test, res


def _makenet(x, num_layers, dropout, random_seed):
    from keras.layers import Dense, Dropout

    dropout_seeder = random.Random(random_seed)

    for i in range(num_layers - 1):
        # add intermediate layers
        if dropout:
            x = Dropout(dropout, seed=dropout_seeder.randint(0, 10000))(x)
        x = Dense(1024, activation="relu", name='dense_layer_{}'.format(i))(x)

    if dropout:
        # add the final dropout layer
        x = Dropout(dropout, seed=dropout_seeder.randint(0, 10000))(x)

    return x

def _save(model, base_model, layers, labels, random_seed, checkpoints_dir):
    from keras.layers import Flatten, Dense
    from keras import Model
    nclasses = len(labels)
    x = Flatten()(base_model.output)
    x = _makenet(x, layers, dropout=None, random_seed=random_seed)
    predictions = Dense(nclasses, activation="softmax", name="predictions")(x)
    model_final = Model(inputs=base_model.input, outputs=predictions)

    for i in range(layers - 1):
        weights = model.get_layer(name='dense_layer_{}'.format(i)).get_weights()
        model_final.get_layer(name='dense_layer_{}'.format(i)).set_weights(weights)

    weights = model.get_layer(name='predictions').get_weights()
    model_final.get_layer(name='predictions').set_weights(weights)

    model_final.save(os.path.join(checkpoints_dir, "model.h5"))
    with open(os.path.join(checkpoints_dir, "labels.txt"), "w") as f:
        f.write("\n".join(labels))
    return model_final


from vergeml.img import INPUT_PATTERNS, open_image, fixext, ImageType
from vergeml.io import source, SourcePlugin, Sample
from vergeml.data import Labels
from vergeml.utils import VergeMLError
from vergeml.sources.labeled_image import LabeledImageSource
import random
import numpy as np
from PIL import Image
import os.path
import json
from operator import methodcaller
import io
from typing import List
import gzip
import hashlib

_FILES = ("train-images-idx3-ubyte.gz", "train-labels-idx1-ubyte.gz",
          "t10k-images-idx3-ubyte.gz", "t10k-labels-idx1-ubyte.gz")


_MNIST_LABELS = ("0", "1", "2", "3", "4", "5", "6", "7", "8", "9")

_FASHION_MNIST_LABELS = ("tshirt_top",
                         "trouser",
                         "pullover",
                         "dress",
                         "coat",
                         "sandal",
                         "shirt",
                         "sneaker",
                         "sag",
                         "ankle_boot")

# we use the md5 to check for fashion mnist, so we can provide the labels
# automatically
_MD5_FASHION = "8d4fb7e6c68d591d4c3dfef9ec88bf0d"


def _md5(fname):
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


@source('image', descr="Load images in MNIST format.")
class InputMnist(SourcePlugin):
    data = None

    def num_samples(self, split: str) -> int:
        return len(self.data[split])

    def read_sample(self, split: str, index: int):
        return self.data[split][index]

    def _check_files(self):
        self.data = dict(train=[], val=[], test=[])

        samples_dir = self.config["samples_dir"]
        files = [os.path.join(samples_dir, file) for file in _FILES]

        for path in files:
            if not os.path.exists(path):
                raise VergeMLError("File not found in samples_dir: {}".format(
                    os.path.basename(path)))

        if _md5(files[0]) == _MD5_FASHION:
            self.meta['labels'] = _FASHION_MNIST_LABELS
        else:
            self.meta['labels'] = _MNIST_LABELS

        # preload
        for split, images, labels in (('train', files[0], files[1]), ('test', files[2], files[3])):

            with gzip.open(images) as f:
                # First 16 bytes are magic_number, n_imgs, n_rows, n_cols
                pixels = np.frombuffer(f.read(), 'B', offset=16)
                pixels = pixels.reshape(-1, 28, 28)

            with gzip.open(labels) as f:
                # First 8 bytes are magic_number, n_labels
                integer_labels = np.frombuffer(f.read(), 'B', offset=8)

            n_cols = integer_labels.max() + 1

            for ix, imagearr in enumerate(pixels):
                label = integer_labels[ix]
                onehot = np.zeros((n_cols), dtype='float32')
                onehot[label] = 1.0
                self.data[split].append((Image.fromarray(imagearr), onehot,
                                         dict(labels=self.meta['labels'],
                                              filename=images,
                                              split=split,
                                              types=('pil', 'labels'))))

            if split == 'train':
                n = self.config['val_num']

                if self.config['val_perc'] is not None:
                    n = int(len(self.data['train']) * self.config['val_perc'] // 100)
                if n is not None:
                    if n > len(self.data['train']):
                        raise VergeMLError("number of test samples is greater than number of available samples.")

                    rng = random.Random(self.config['random_seed'])
                    count = len(self.data[split])
                    indices = rng.sample(range(count), count)
                    self.data['val'] = [self.data['train'][i] for i in indices[:n]]
                    self.data['train'] = [self.data['train'][i] for i in indices[n:]]

            else:

                if self.config['test_num']:
                    if self.config['test_num'] > len(self.data['test']):
                        raise VergeMLError("number of test samples is greater than number of available samples.")

                    rng = random.Random(self.config['random_seed'])
                    indices = rng.sample(range(len(self.data[split])), len(pixels))
                    self.data['test'] = [self.data['test'][i] for i in indices[:n]]


plugin = InputMnist



from vergeml.utils import VergeMLError
from vergeml.dataset import DatasetPlugin, dataset
from vergeml.display import DISPLAY
import shutil
import os.path
import zipfile
import tarfile
import json
import urllib.parse
import glob
import csv

_LONG_DESCR = """
A Large Collection of Dermatoscopic Images of Common Pigmented Skin Lesions

The "Human Against Machine with 10000 training images" dataset is a collection of
dermatoscopic images from different populations acquired and stored by different 
modalities. It was created by applying different acquisition and cleaning methods.
The final dataset consists of 10010 dermatoscopic images. This benchmark dataset 
can be used for machine learning and for comparisons with human experts. Cases 
include a representative collection of all important diagnostic categories in the 
realm of pigmented lesions. More than 50% of lesions have been confirmed by pathology, 
the ground truth for the rest of the cases was either follow-up, expert consensus, 
or confirmation by in-vivo confocal microscopy.
""".strip()

_INFO = [
    ['Samples', '10K'],
    ['Type', 'Labeled Images'],
    ['Size', '2.5G']
]

_LONG_DESCR += "\n\n" + DISPLAY.table(_INFO, separate='none').getvalue()

_URL = "https://dataverse.harvard.edu/dataset.xhtml?persistentId=doi:10.7910/DVN/DBW86T"
_URL1 = "https://dataverse.harvard.edu/api/access/datafile/3172585?gbrecs=true"
_URL2 = "https://dataverse.harvard.edu/api/access/datafile/3172584?gbrecs=true"
_URL3 = "https://dataverse.harvard.edu/api/access/datafile/3172582?format=original&gbrecs=true"


@dataset('ham10000', descr="Dermatoscopic images of skin lesions.", long_descr=_LONG_DESCR)
class Ham10KDataset(DatasetPlugin):
    
    def __call__(self, args, env):
        samples_dir = env.get('samples-dir')

        print("Downloading HAM10000 to {}.".format(samples_dir))

        jsessionid = None
        fp = urllib.request.urlopen(_URL)
        info = fp.info()
        fp.close()

        cookies = info.get("Set-Cookie")

        for cookie in cookies.split(";"):
            if cookie.startswith("JSESSIONID"):
                _, jsessionid = cookie.split("=")
                jsessionid = jsessionid.strip()
                break

        if not jsessionid:
            raise VergeMLError("Could not get jsessionid.")
        
        files = [(_URL1, "ham10000_1.zip"), (_URL2, "ham10000_2.zip"), (_URL3, "meta.csv")]
     
        src_dir = self.download_files(files, env=env, headers=[('Cookie', cookies)], dir=env.get('cache-dir'))

        print("")
        print("Extracting data...")
        for fname in ("ham10000_1.zip", "ham10000_2.zip"):
            path = os.path.join(src_dir, fname)
            zipf = zipfile.ZipFile(path, 'r')
            zipf.extractall(src_dir)
            zipf.close()
        
        meta_path = os.path.join(src_dir, "meta.csv")
        
        with open(meta_path, 'r') as csvfile:
            csv_reader = csv.reader(csvfile)
            header = True
            
            for row in csv_reader:
                if header:
                    header = False
                    continue
                [_lesion_id, image_id, dx, _dx_type, _age, _sex, _localization] = row

                dirname = os.path.join(samples_dir, dx)
                if not os.path.exists(dirname):
                    os.makedirs(dirname)

                shutil.copy(os.path.join(src_dir, image_id + ".jpg"), os.path.join(samples_dir, dx, image_id + ".jpg"))
        
        shutil.copy(os.path.join(src_dir, "meta.csv"), os.path.join(samples_dir, "meta.csv"))
        
        shutil.rmtree(src_dir)
        print("Finished downloading HAM10000.")



from vergeml.utils import VergeMLError
from vergeml.dataset import DatasetPlugin, dataset
from vergeml.display import DISPLAY
import shutil
import os.path
import zipfile
import tarfile

_LONG_DESCR = """
25K images of cats and dogs.

Credits:
  Jeremy Elson, John R. Douceur, Jon Howell, Jared Saul, Asirra:
  A CAPTCHA that Exploits Interest-Aligned Manual Image
  Categorization, in Proceedings of 14th ACM Conference on Computer
  and Communications Security (CCS), Association for Computing
  Machinery, Inc., Oct. 2007

For more information visit https://www.kaggle.com/c/dogs-vs-cats
""".strip()

_INFO = [
    ['Samples', '25K'],
    ['Type', 'Labeled Images'],
    ['Size', '786.7 MB']
]

_LONG_DESCR += "\n\n" + DISPLAY.table(_INFO, separate='none').getvalue()

_URL = "https://download.microsoft.com/download/3/E/1/3E1C3F21-ECDB-4869-8368-6DEBA77B919F/kagglecatsanddogs_3367a.zip"

@dataset('cats-and-dogs', descr="25K images of cats and dogs.", long_descr=_LONG_DESCR)
class CatsAndDogsDataset(DatasetPlugin):

    def __call__(self, args, env):
        samples_dir = env.get('samples-dir')
        for label in ("cat", "dog"):
            dest = os.path.join(samples_dir, label)
            if os.path.exists(dest):
                raise VergeMLError("Directory {} already exists in samples dir: {}".format(label, dest))
        print("Downloading cats and dogs to {}.".format(samples_dir))
        src_dir = self.download_files([(_URL, "catsdogs.zip")], env)
        path = os.path.join(src_dir, "catsdogs.zip")

        print("Extracting data.")
        zipf = zipfile.ZipFile(path, 'r')
        zipf.extractall(src_dir)
        zipf.close()

        for file, dest in (("PetImages/Dog", "dog"), ("PetImages/Cat", "cat")):
            shutil.copytree(os.path.join(src_dir, file), os.path.join(samples_dir, dest))

        shutil.rmtree(src_dir)

        # WTF?
        os.unlink(os.path.join(samples_dir, "cat", "666.jpg"))
        os.unlink(os.path.join(samples_dir, "dog", "11702.jpg"))

        print("Finished downloading cats and dogs.")



"""List command.
"""
from copy import deepcopy
from collections import OrderedDict
import os
import os.path
import json
import csv
from datetime import datetime
import io
import sys

import yaml

from vergeml.command import command, CommandPlugin, Command
from vergeml.option import option
from vergeml.utils import VergeMLError
from vergeml.display import DISPLAY
from vergeml.config import parse_command

EXAMPLES = """
$ ml list -sacc
# sort by acc value

$ ml list status -eq RUNNING
# show trainings that are currently running

$ ml list test_acc -gt 0.8
# show AIs with a test accuracy that is greater than 0.8

# available comparison operations:
# -gt, -lt, -eq, -neq, -gte and -lte
""".strip()

@command('list', descr="List trained models.", free_form=True, examples=EXAMPLES) # pylint: disable=R0903
@option('sort', descr="By which column to sort.", default='created-at', short='s')
@option('order', descr="Sort order.", default='asc', short='o', validate=('asc', 'desc'))
@option('columns', descr="Which columns to show.", type='Optional[Union[str, List[str]]]', short='c')
@option('output', descr="Output format.", default='table', validate=('table', 'csv', 'json'))
class ListCommand(CommandPlugin):
    """List trained models"""

    def __call__(self, args, env):

        # Parse and partition into normal and comparison args.
        args, cargs = _parse_args(args, env)

        # When trainings dir does not exist, print an error and exit
        if not os.path.exists(env.get('trainings-dir')):
            print("No trainings found.", file=sys.stderr)
            return

        info, hyper = _find_trained_models(args, env)

        theader, tdata, left_align = _format_table(args, cargs, info, hyper)

        _output_table(args['output'], theader, tdata, left_align)

def _parse_args(args, env):
    args = args[1]

    comps = []
    for idx, arg in enumerate(args):
        if arg in ('-gt', '-lt', '-eq', '-neq', '-gte', '-lte'):
            start, end = idx - 1, idx + 1
            if start < 0 or end >= len(args):
                raise VergeMLError("Invalid options.", help_topic='list')
            comps.append((start, end))

    cargs = []
    for start, end in reversed(comps):
        cargs.append(args[start:end+1])
        del args[start:end+1]

    cmd = deepcopy(Command.discover(ListCommand))
    cmd.free_form = False
    args.insert(0, 'list')
    args = cmd.parse(args)

    # If existent, read settings from the config file
    config = parse_command(cmd, env.get(cmd.name))

    # Set missing args from the config file
    for k, arg in config.items():
        args.setdefault(k, arg)

    # Set missing args from default
    for opt in cmd.options:
        if opt.name not in args and (opt.default is not None or not opt.is_required()):
            args[opt.name] = opt.default

    return args, cargs

def _find_trained_models(args, env):
    info = {}
    hyper = {}
    train_dir = env.get('trainings-dir')

    for trained_model in os.listdir(train_dir):
        data_yaml = os.path.join(train_dir, trained_model, 'data.yaml')
        if os.path.isfile(data_yaml):
            with open(data_yaml) as file:
                doc = yaml.safe_load(file)
        else:
            doc = {}
        info[trained_model] = {}
        hyper[trained_model] = {}

        if 'model' in doc:
            info[trained_model]['model'] = doc['model']

        if 'results' in doc:
            info[trained_model].update(doc['results'])

        if 'hyperparameters' in doc:
            hyper[trained_model].update(doc['hyperparameters'])

    sort = [s.strip() for s in args['sort'].split(",")]

    info = OrderedDict(sorted(info.items(), reverse=(args['order'] == 'asc'),
                              key=lambda x: [x[1].get(s, 0) for s in sort]))
    return info, hyper

def _format_table(args, cargs, info, hyper): # pylint: disable=R0912
    theader = ['AI', 'model', 'status', 'num-samples', 'training-start', 'epochs']
    exclude = ['training-end', 'steps', 'created-at']

    if args['columns']:
        cols = args['columns']
        if isinstance(cols, str):
            cols = cols.split(",")

        theader = ['AI'] + [s.strip() for s in cols]
        exclude = []

    tdata = []
    left_align = set([0])

    for trained_model, results in info.items():
        rdata = [""] * len(theader)
        rdata[0] = "@" + trained_model

        if not _filter(results, hyper[trained_model], cargs):
            continue

        for k, val in sorted(results.items()):
            if k in exclude and not args['columns']:
                continue

            if not k in theader and not args['columns'] and isinstance(val, (str, int, float)):
                theader.append(k)
                rdata.append(None)

            if k in theader:
                pos = theader.index(k)

                if k in ('training-start', 'training-end', 'created-at'):
                    val = datetime.utcfromtimestamp(val)
                    val = val.strftime("%Y-%m-%d %H:%M")
                elif isinstance(val, float):
                    val = "%.4f" % val
                elif isinstance(val, str):
                    left_align.add(pos)

                rdata[pos] = val

        for k, val in sorted(hyper[trained_model].items()):

            if k in theader:
                pos = theader.index(k)
                if isinstance(val, float):
                    val = "%.4f" % val
                elif isinstance(val, str):
                    left_align.add(pos)

                rdata[pos] = val

        tdata.append(rdata)

    return theader, tdata, left_align

def _output_table(output, theader, tdata, left_align):

    if not tdata:
        print("No matching trained models found.", file=sys.stderr)

    if output == 'table':
        if not tdata:
            return
        tdata.insert(0, theader)
        print(DISPLAY.table(tdata, left_align=left_align).getvalue(fit=True))

    elif output == 'json':
        res = []
        for row in tdata:
            res.append(dict(zip(theader, row)))
        print(json.dumps(res))

    elif output == 'csv':
        buffer = io.StringIO()

        writer = csv.writer(buffer)
        writer.writerow(theader)
        for row in tdata:
            writer.writerow(row)
        val = buffer.getvalue()
        val = val.replace('\r', '')
        print(val.strip())

def _filter(info, hyper, comp_args):
    try:
        cols = {}
        cols.update(hyper)
        cols.update(info)
        res = True
        for col, opr, val in comp_args:

            if not col in cols:
                return False

            cval = cols[col]

            if isinstance(cval, int):
                val = int(val)
            elif isinstance(cval, float):
                val = float(val)

            if opr == '-eq':
                res = res and (cval == val)
            elif opr == '-neq':
                res = res and (cval != val)
            elif opr == '-gt':
                res = res and (cval > val)
            elif opr == '-lt':
                res = res and (cval < val)
            elif opr == '-gte':
                res = res and (cval >= val)
            elif opr == '-lte':
                res = res and (cval <= val)

            if not res:
                return False
        return res
    except: # pylint: disable=W0702
        return False



from typing import List, Generator, Any, Tuple, Union
import operator
import random
from vergeml.io import Sample
from vergeml.utils import VergeMLError, SPLITS
from copy import copy
from vergeml.option import option, Option
from vergeml.plugins import PLUGINS

_OPERATION_META_KEY = '__vergeml_operation__'

# TODO should all operations take the apply parameter?
def operation(name, descr=None, long_descr=None, topic='general', apply=True):
    """Define an operation.

    :param name:        Name of the operation.
    :param descr:       A short description of the operation
    :param long_descr:  A long description
    :param apply:       If the preprocessing operation takes the apply parameter.
    """
    def decorator(o):
        assert getattr(o, _OPERATION_META_KEY, None) is None

        options = Option.discover(o)
        cmd = Operation(name,
                        descr=descr,
                        long_descr=long_descr,
                        apply=apply,
                        options=options,
                        topic=topic)
        setattr(o, _OPERATION_META_KEY, cmd)
        return o
    return decorator

class Operation:
    def __init__(self, name, descr=None, long_descr=None, apply=True, topic='general', options=[], plugins=PLUGINS):
        self.name = name
        self.descr = descr
        self.long_descr = long_descr
        self.apply = apply
        self.options = options
        self.plugins = plugins
        self.topic = topic

    @staticmethod
    def discover(o, plugins=PLUGINS):
        res = None
        if hasattr(o, _OPERATION_META_KEY):
            res = getattr(o, _OPERATION_META_KEY)
            res.plugins = plugins
            for option in res.options:
                option.plugins = plugins
        return res


class BaseOperation:
    """Base class for processing samples.

    This class can be used to augment, filter, transform and combine samples.
    To support this variety of use-cases, BaseOperation offers fine-grained control
    over how samples are being processed.

    When an operation changes the number of output samples, it must return the
    factor in multiplier().
    """

    def configuration(self):
        """Return the configuration of the BaseOperation instance.

        Since operations play a role in data processing, any change in a operation
        pipeline must result in a different hash value for a dataset. To capture the
        configuration of a preprocessing operation, BaseOperation defines this method,
        which can be overridden when needed.
        """
        return self.__dict__

    def process(self, sample: Sample, ops=List['BaseOperation']) -> Generator[Sample, None, None]:
        """Complete a processing step in the pipeline and run the next one.

        :param sample: The sample to be processed
        :param ops: The next operations to run

        :return: A generator yielding samples

        The process function is expected to first transform the sample and then to call
        the next BaseOperation, yielding the resulting sample as a return value.
        """
        raise NotImplementedError

    def multiplier(self) -> float:
        """Return the factor by which the operation changes the number of output samples"""
        return 1.0


class OperationPlugin(BaseOperation):
    """Simplified Operations.

    Most operations will not need the raw power of BaseOperation, so an easier to use
    class is provided. In addition to simplified processing, it also offers the following
    functionality:

    - control which split the operation is applied to
    - TODO control which labels the operation is applied to
    - control if the operation is applied to the samples or the ground truth or both.
    - Automatic type checking so that an operation is only applied to supported sample data.
    """

    # To implement  type checking functionality, a subclass has to set this field to the type it
    # can handle. If it can handle multiple types, Union[Type1, Type2, ..] can be used
    type = Any

    # apply can be set from the operation configuration. it can be a comma separated string, a
    # tuple or None.
    apply = None

    def __init__(self, apply=None):
        """
        :param apply: a tuple, a comma separated string or None
                      possible values are a combination of:
                        - x: apply the operation to the sample only
                        - y: apply the operation to the label only
                        - train: apply the operation only to the train split
                        - val: apply the operation only to the val split
                        - test: apply the operation only to the test split
                        - all: apply the operation to all
        """
        super().__init__()

        assert isinstance(apply, (list, tuple, str, type(None)))

        if apply == 'all':
            self.apply = set()
        elif isinstance(apply, (list, tuple)):
            self.apply = set(apply)
        elif isinstance(apply, str):
            self.apply = set(map(operator.methodcaller('strip'), apply.split(",")))
        else:
            self.apply = set()
            # CORRECT:
            # self.apply = {'train'}

    def transform(self, data: Any, rng: random.Random) -> Any:
        """Transform either x or y.

        :param data: the data to transform
        :param rng: the random generator

        :return: the transformed data

        transform() will pass in data, which can be either x or the y.
        When x and y of the same sample are processed with transform, the random number
        generator will be reset on the second call so that both pieces of data are
        processed with the same random numbers.
        """
        raise NotImplementedError

    def transform_xy(self, x: Any, y: Any, rng: random.Random) -> Tuple[Any, Any]:
        """Transform x and y values.

        :param x: the sample data
        :param y: the ground truth
        :param rng: random generator

        :return: a tuple (x, y)

        When you need to transform x and y values in the same method, for example
        when they are images with different sizes and they need to be resized to have
        the same size, override this method.
        """
        rngstate = rng.getstate()
        appxy = self.apply.intersection({'x', 'y'})

        if not appxy or 'x' in self.apply:
            if _type_good(self.type, type(x)):
                rng.setstate(rngstate)
                x = self.transform(x, rng)

        if not appxy or 'y' in self.apply:
            if _type_good(self.type, type(y)):
                rng.setstate(rngstate)
                y = self.transform(y, rng)

        return x, y

    def transform_sample(self, sample: Sample) -> Sample:
        """Transform a sample.
        :param sample: the sample to transform

        :return: the transformed sample:

        When you don't care about the apply functionality and just want to process the
        sample, override this method.
        """
        if self.apply.intersection(set(SPLITS)) \
                and sample.meta['split'] not in self.apply:
            yield sample
        else:
            x, y = self.transform_xy(sample.x, sample.y, sample.rng)
            yield Sample(x, y, sample.meta, sample.rng)

    def process(self, sample: Sample, ops=List[BaseOperation]) -> Generator[Sample, None, None]:

        for s1 in self.transform_sample(sample):
            if not ops:
                yield s1
            else:
                nextop, *rest = ops
                yield from nextop.process(s1, rest) # pylint: disable=E1101

    def configuration(self):
        # CORRECT:
        # res = self.__dict__.copy()
        # if 'apply' in res:
        #     res['apply'] = list(sorted(res['apply']))
        return self.__dict__.copy()


def _type_good(t1, t2):
    return t1 == Any or t1 == t2 or \
           (getattr(t1, '__origin__', None) == Union and t2 in t1.__args__)
